from abc import abstractmethod
from typing import Union, TYPE_CHECKING, Tuple, Dict, Any

from numpy import zeros

from mewpy.germ.solution import ModelSolution
from mewpy.solvers.solution import Solution
from mewpy.solvers.solver import Solver
from .linear_containers import ConstraintContainer, VariableContainer
from .linear_utils import LinkedList, Node, get_solver_instance

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


class LinearProblem:

    def __init__(self,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 solver: Union[str, Solver] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        Linear programing base implementation. A GERM model is converted into a linear problem using reframed/mewpy
        solver interface. Both CPLEX and Gurobi solvers are currently supported. Other solvers may also be supported
        using an additional OptLang solver interface. However, CPLEX and Gurobi are recommended for certain problems.

        A linear problem is linked with a given model via asynchronous updates.
        That is, alterations to the model are sent to all attached simulators via notification objects.
        Notifications are processed accordingly by all linear problems attached to the model.
        Each implementation of a linear problem (e.g. FBA, RFBA, SRFBA, etc) is responsible
        for processing the notifications in the correct way.

        A linear problem has one and only one solver object.
        Alterations to a linear problem are promptly forced in the solver by building a new solver instance.
        Alternatively, one can impose temporary constraints during problem optimization
        (see the method for further details)

        Notes for developers:
        A linear problem object is an observer (observer pattern) of a germ model.
        A notification with model changes is sent to all observers (linear problems).
        The linear problem implementation specific for each method processes the notification accordingly.
        Finally, when the linear problem is updated,
        all variables and constraints added to the linear problem are implemented and kept in sync with the solver
        This can avoid consecutive building of the solver, namely a lazy loading

        :param model: a MetabolicModel, RegulatoryModel or GERM model. The model is used to retrieve
        variables and constraints to the linear problem
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used,
        but it will be overwritten if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        if not model:
            raise ValueError('A valid model must be provided')

        self._model = model

        # this is useful to restore the linear problem to the point of init
        self._initial_solver = solver
        self._solver = None
        self._synchronized = False

        # Simulator index engine uses a linked list with a built-in dictionary. This allows fast access to the index
        # of a given variable or constraint.
        # Note that, some variables or constraints can comprise multiple rows or columns,
        # so that the job of keeping in track of all indexes of a given variable/constraint is actually way
        # harder than it seems for simple linear problems (e.g. fba)
        self._cols = LinkedList()
        self._rows = LinkedList()

        # one to one indexing of all variables
        self._sub_cols = LinkedList()

        # Holding constraints and variables objects
        self._constraints = {}
        self._variables = {}

        # the objective is a dict variable_id: coefficient
        self._linear_objective = {}
        self._quadratic_objective = {}
        self._minimize = True

        if build:
            self.build()

        if attach:
            self.model.attach(self)

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):
        return f"{self.method} for {self.model.id}"

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        """
        It returns a html representation of the linear problem
        :return:
        """
        if self.solver:
            solver = self.solver.__class__.__name__
        else:
            solver = 'None'

        return f"""
        <table>
            <tr>
                <td>Method</td>
                <td>{self.method}</td>
            </tr>
            <tr>
                <td>Model</td>
                <td>{self.model}</td>
            </tr>
            <tr>
                <th>Variables</th>
                <td>{len(self.variables)}</td>
            </tr>
            <tr>
                <th>Constraints</th>
                <td>{len(self.constraints)}</td>
            </tr>
            <tr>
                <th>Objective</th>
                <td>{self.objective}</td>
            </tr>
            <tr>
                <th>Solver</th>
                <td>{solver}</td>
            </tr>
            <tr>
                <th>Synchronized</th>
                <td>{self.synchronized}</td>
            </tr>
        </table>
        """

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @property
    def method(self) -> str:
        """
        Name of the method implementation to build and solve the linear problem
        :return: the name of the class
        """
        return self.__class__.__name__

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        """
        GERM model of this simulator
        :return: a MetabolicModel, RegulatoryModel or GERM model
        """
        return self._model

    @property
    def solver(self) -> Solver:
        """
        mewpy solver instance for this linear problem. It contains an interface for the concrete solver
        :return: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance
        """
        return self._solver

    @property
    def synchronized(self) -> bool:
        """
        Whether the linear problem is synchronized with the model
        :return:
        """
        return self._synchronized

    @property
    def constraints(self) -> Dict[str, ConstraintContainer]:
        """
        A copy of the constraints' container.
        This container holds all ConstraintContainer objects for this linear problem.
        Note that, a constraint container can hold several constraints/rows
        :return: copy of the constraints dictionary
        """
        return self._constraints.copy()

    @property
    def variables(self) -> Dict[str, VariableContainer]:
        """
        A copy of the variables' container.
        This container holds all VariableContainer objects for this linear problem.
        Note that, a variable container can hold several variables/columns
        :return: copy of the variables dictionary
        """
        return self._variables.copy()

    @property
    def objective(self) -> Dict[Union[str, Tuple[str, str]], Union[float, int]]:
        """
        A copy of the objective dictionary. Keys are either variable identifiers or tuple of variable identifiers.
        Values are the corresponding coefficients
        Note that, linear and quadratic objectives can be encoded in the objective dictionary.
        See the set_objective method for further detail
        :return: copy of the objective dictionary
        """
        return {**self._linear_objective, **self._quadratic_objective}

    @property
    def minimize(self) -> bool:
        """
        The linear problem objective sense/direction
        :return: a boolean whether the linear problem objective sense/direction is minimization
        """
        return bool(self._minimize)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def matrix(self):
        """
        The linear problem matrix
        :return: a matrix as numpy array
        """
        return self._get_matrix()

    @property
    def bounds(self):
        """
        The linear problem bounds
        :return: bounds as list of tuples
        """
        return self.get_bounds(as_list=True)

    @property
    def b_bounds(self):
        """
        The linear problem b bounds (constraints bounds)
        :return: b bounds as list of tuples
        """
        return self.get_bounds(b_bounds=True, as_list=True)

    @property
    def shape(self):
        """
        The linear problem shape
        :return: a tuple with the number of rows and columns
        """
        return int(len(self._rows)), int(len(self._cols))

    # -----------------------------------------------------------------------------
    # MEWpy solver
    # -----------------------------------------------------------------------------
    def build_solver(self, variables: bool = True, constraints: bool = True, objective: bool = True):
        """
        It creates a new solver instance and adds the current state (variables, constraints) of the linear problem
        to the solver.
        :param variables: Whether to add variables to the solver. Default: True
        :param constraints: Whether to add constraints to the solver. Default: True
        :param objective: Whether to add the objective to the solver. Default: True
        :return:
        """
        if variables or constraints:
            self._solver = get_solver_instance(self._initial_solver)

        if variables:
            for variable in self._variables.values():

                # Using mewpy/reframed solver interface ...
                for name, (lb, ub, var_type) in variable.items():
                    self.solver.add_variable(var_id=name, lb=lb, ub=ub, vartype=var_type, update=False)

            self.solver.update()

        if constraints:
            for i, constraint in enumerate(self._constraints.values()):

                # Using mewpy/reframed solver interface ...
                for j, (coef, lb, ub) in constraint.items():

                    cnt_id = str(i + j)

                    if lb == ub:
                        rhs = lb
                        self.solver.add_constraint(constr_id=cnt_id, lhs=coef, sense='=', rhs=rhs, update=False)

                    else:
                        cnt_id_f = f'{cnt_id}_forward'
                        rhs = lb
                        self.solver.add_constraint(constr_id=cnt_id_f, lhs=coef, sense='>', rhs=rhs, update=False)

                        cnt_id_r = f'{cnt_id}_reverse'
                        rhs = ub
                        self.solver.add_constraint(constr_id=cnt_id_r, lhs=coef, sense='<', rhs=rhs, update=False)

            self.solver.update()

        if objective:
            linear_objective = {}

            if self._linear_objective:

                for k, v in self._linear_objective.items():

                    if k not in self._cols and k not in self._sub_cols:
                        raise ValueError(f'{k} is not a variable of this linear problem')

                    linear_objective[k] = v

            quadratic_objective = {}

            if self._quadratic_objective:

                for (k1, k2), v in self._quadratic_objective.items():

                    if k1 not in self._cols and k1 not in self._sub_cols:
                        raise ValueError(f'{k1} is not a variable of this linear problem')

                    if k2 not in self._cols and k2 not in self._sub_cols:
                        raise ValueError(f'{k2} is not a variable of this linear problem')

                    quadratic_objective[(k1, k2)] = v

            self.solver.set_objective(linear_objective, quadratic_objective, self._minimize)
            self.solver.update()

    # -----------------------------------------------------------------------------
    # Clean
    # -----------------------------------------------------------------------------
    def clean(self):
        """
        It cleans the linear problem object by removing all variables and constraints
        :return:
        """
        self._synchronized = False
        self._solver = get_solver_instance(self._initial_solver)
        self._cols = LinkedList()
        self._rows = LinkedList()
        self._sub_cols = LinkedList()
        self._constraints = {}
        self._variables = {}
        self._linear_objective = {}
        self._quadratic_objective = {}
        self._minimize = True
        return

    # -----------------------------------------------------------------------------
    # Build
    # -----------------------------------------------------------------------------
    @abstractmethod
    def _build(self):
        """
        Abstract method for the concrete build method
        :return:
        """
        pass

    def build(self) -> 'LinearProblem':
        """
        Abstract implementation
        :return:
        """
        # clean first
        self.clean()

        # concrete build
        self._build()

        # build solver
        self.build_solver(variables=True, constraints=True, objective=True)

        # update status
        self._synchronized = True
        return self

    # -----------------------------------------------------------------------------
    # Optimization
    # -----------------------------------------------------------------------------
    @abstractmethod
    def _optimize(self, solver_kwargs: Dict[str, Any] = None, **kwargs) -> Solution:
        """
        Abstract method for the concrete optimization method
        :param solver_kwargs: solver specific keyword arguments
        :param kwargs: keyword arguments
        :return:
        """
        pass

    def optimize(self,
                 to_solver: bool = False,
                 solver_kwargs: Dict[str, Any] = None,
                 **kwargs) -> Union[ModelSolution, Solution]:
        """
        It solves the linear problem. The linear problem is solved using the solver interface.

        The optimize method allows setting temporary changes to the linear problem. The changes are
        applied to the linear problem reverted to the original state afterward.
        Objective, constraints and solver parameters can be set temporarily.

        The solution is returned as a ModelSolution instance, unless to_solver is True. In this case,
        the solution is returned as a SolverSolution instance.

        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False.
        Otherwise, a ModelSolution is returned.
        :param solver_kwargs: Solver parameters to be set temporarily.
            - linear: A dictionary of linear coefficients to be set temporarily. The keys are the variable names
            and the values are the coefficients. Default: None
            - quadratic: A dictionary of quadratic coefficients to be set temporarily. The keys are tuples of
            variable names and the values are the coefficients. Default: None
            - minimize: Whether to minimize the objective. Default: False
            - constraints: A dictionary with the constraints bounds. The keys are the constraint ids and the values
            are tuples with the lower and upper bounds. Default: None
            - get_values: Whether to retrieve the solution values. Default: True
            - shadow_prices: Whether to retrieve the shadow prices. Default: False
            - reduced_costs: Whether to retrieve the reduced costs. Default: False
            - pool_size: The size of the solution pool. Default: 0
            - pool_gap: The gap between the best solution and the worst solution in the pool. Default: None
        :return: A ModelSolution instance or a SolverSolution instance if to_solver is True.
        """
        # build solver if out of sync
        if not self.synchronized:
            self.build()

        if not solver_kwargs:
            solver_kwargs = {}

        # concrete optimize
        solution = self._optimize(solver_kwargs=solver_kwargs, **kwargs)

        if to_solver:
            return solution

        minimize = solver_kwargs.get('minimize', self._minimize)
        return ModelSolution.from_solver(method=self.method, solution=solution, model=self.model,
                                         minimize=minimize)

    # -----------------------------------------------------------------------------
    # Update - Observer interface
    # -----------------------------------------------------------------------------
    def update(self):
        """
        It updates the linear problem object by adding/removing variables and constraints
        Note that linear problems are not updated after each addition/removal of a variable or constraint to the model.
        This is done to avoid unnecessary updates of the solver. Instead, the update is done when the method
        `build` is called. If required, this method is called by the simulation methods (e.g. fba, pfba, etc) before the
        optimization process in the `optimize` method.
        :return:
        """
        self._synchronized = False

    # -----------------------------------------------------------------------------
    # Objective
    # -----------------------------------------------------------------------------
    def set_objective(self,
                      linear: Union[str, Dict[str, Union[float, int]]] = None,
                      quadratic: Dict[Tuple[str, str], Union[float, int]] = None,
                      minimize: bool = True):
        """
        A dictionary of the objective for the linear problem.
        Keys must be variables of the linear problem,
        whereas values must be the corresponding coefficients as int or float.
        It can be changed during optimization

        :param linear: a dictionary of linear coefficients or variable identifier (that is set with a coefficient of 1)
        :param quadratic: a dictionary of quadratic coefficients.
        Note that keys must be a tuple of reaction pairs to be summed up to a quadratic objective function
        :param minimize: whether to solve a minimization problem. This parameter is True by default
        :return:
        """
        if linear is None:
            linear = {}

        if quadratic is None:
            quadratic = {}

        if isinstance(linear, str):
            linear = {linear: 1}

        if not isinstance(linear, dict):
            raise TypeError(f'linear objective must be a dictionary, not {type(linear)}')

        if not isinstance(quadratic, dict):
            raise TypeError(f'quadratic objective must be a dictionary, not {type(quadratic)}')

        if not isinstance(minimize, bool):
            raise TypeError(f'minimize must be a boolean, not {type(minimize)}')

        self._linear_objective = linear
        self._quadratic_objective = quadratic
        self._minimize = minimize
        self.build_solver(objective=True)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations - add/remove variables and constraints
    # -----------------------------------------------------------------------------
    def add_constraints(self, *constraints: ConstraintContainer):
        for constraint in constraints:
            if constraint.name in self._rows:
                # The constraint is replaced, as the linear problem behaves like a set
                # This also mimics the solver interface behavior

                old_constraint = self._constraints[constraint.name]

                self.remove_constraints(old_constraint)

            node = constraint.to_node()

            self._rows.add(node)

            self._update_constraint_coefs(constraint)

            self._constraints[constraint.name] = constraint

    def remove_constraints(self, *constraints: ConstraintContainer):
        for constraint in constraints:
            if constraint.name in self._rows:
                self._rows.pop(constraint.name)
                self._constraints.pop(constraint.name)

    def add_variables(self, *variables: VariableContainer):
        for variable in variables:
            if variable.name in self._cols:
                # The variable is replaced, as the linear problem behaves like a set
                # This also mimics the solver interface behavior

                old_variable = self._variables[variable.name]

                self.remove_variables(old_variable)

            node = variable.to_node()

            self._cols.add(node)

            self._variables[variable.name] = variable

            for sub_variable in variable.keys():
                sub_node = Node(value=sub_variable, length=1)

                self._sub_cols.add(sub_node)

    def remove_variables(self, *variables: VariableContainer):
        for variable in variables:
            if variable.name in self._cols:

                self._cols.pop(variable.name)
                self._variables.pop(variable.name)

                for sub_variable in variable.keys():
                    self._sub_cols.pop(sub_variable)

    def _update_constraint_coefs(self, constraint: ConstraintContainer):

        # some constraint coefficients might have keys that refer to the variable name and not sub-variable name.
        # Since only sub-variable names are added to the solvers,
        # these keys must be updated to the last sub-variable name. Note that, the last sub-variable name is regularly
        # the one that matters, as the initial sub-variables regularly decide the outcome of the last one.

        new_coefs = []

        for coefficient in constraint:

            new_coef = {}

            for var, coef in coefficient.items():

                if var in self._sub_cols:

                    sub_var = var

                else:

                    variable = self._variables[var]

                    sub_var = variable.sub_variables[-1]

                new_coef[sub_var] = coef

            new_coefs.append(new_coef)

        constraint.coefs = new_coefs

    # -----------------------------------------------------------------------------
    # Getters
    # -----------------------------------------------------------------------------
    def index(self, variable=None, constraint=None, as_list=False, as_int=False, default=None):
        """
        It returns the index of a variable or constraint
        :param variable: a variable container
        :param constraint: a constraint container
        :param as_list: a boolean indicating whether the index should be returned as a list
        :param as_int: a boolean indicating whether the index should be returned as an integer
        :param default: a default value to be returned if the variable or constraint is not found
        :return: the index of the variable or constraint
        """
        if variable is None and constraint is None:
            raise ValueError('Please provide a variable or constraint')

        if constraint is not None:

            slc = self._rows.get(constraint)

        else:

            slc = self._cols.get(variable, self._sub_cols.get(variable))

            if slc is None:
                return default

        if as_list:
            return [i for i in range(slc.start, slc.stop)]

        elif as_int:
            return slc.stop - 1

        else:

            return slc

    def _get_matrix(self):

        matrix = zeros(self.shape)

        n = 0
        for cnt in self._constraints.values():

            for coef in cnt.coefs:

                for var, value in coef.items():

                    m = self.index(variable=var, as_int=True)

                    if m is None:
                        continue

                    matrix[n, m] = value

                n += 1

        return matrix

    def _get_b_bounds(self, as_list=False, as_tuples=False):

        if as_list:

            b_bounds = []

            for cnt in self._constraints.values():
                bds = list(zip(cnt.lbs, cnt.ubs))

                b_bounds.extend(bds)

            return b_bounds

        elif as_tuples:

            lbs = []
            ubs = []

            for cnt in self._constraints.values():
                lbs.extend(cnt.lbs)
                ubs.extend(cnt.ubs)

            return tuple(lbs), tuple(ubs)

        else:
            return {key: (cnt.lbs, cnt.ubs) for key, cnt in self._constraints.items()}

    def _get_bounds(self, as_list=False, as_tuples=False):

        if as_list:

            bounds = []

            for var in self._variables.values():
                bds = list(zip(var.lbs, var.ubs))

                bounds.extend(bds)

            return bounds

        elif as_tuples:

            lbs = []
            ubs = []

            for var in self._variables.values():
                lbs.extend(var.lbs)
                ubs.extend(var.ubs)

            return tuple(lbs), tuple(ubs)

        else:
            return {key: (var.lbs, var.ubs) for key, var in self._variables.items()}

    def _get_variable_bounds(self, variable):

        variable = self._variables.get(variable)

        if variable is None:
            lb, ub = self._get_bounds(as_list=True)[self._sub_cols.get(variable)]

            return [lb], [ub]

        return variable.lbs, variable.ubs

    def _get_constraint_bounds(self, constraint):

        constraint = self._constraints.get(constraint)

        return constraint.lbs, constraint.ubs

    def get_bounds(self,
                   variable=None,
                   constraint=None,
                   b_bounds=False,
                   as_list=False,
                   as_tuples=False):
        """
        It returns the bounds of a variable or constraint
        :param variable: a variable container
        :param constraint: a constraint container
        :param b_bounds: a boolean indicating whether the bounds of the constraints should be returned
        :param as_list: a boolean indicating whether the bounds should be returned as a list
        :param as_tuples: a boolean indicating whether the bounds should be returned as a tuple
        :return: the bounds of the variable or constraint
        """
        if variable is not None:

            return self._get_variable_bounds(variable=variable)

        elif constraint is not None:

            return self._get_constraint_bounds(constraint)

        elif b_bounds:

            return self._get_b_bounds(as_list=as_list, as_tuples=as_tuples)

        else:
            return self._get_bounds(as_list=as_list, as_tuples=as_tuples)



from collections import defaultdict
from typing import Union, TYPE_CHECKING, Dict, Tuple, Optional, Sequence

import pandas as pd

from mewpy.util.constants import ModelConstants
from .analysis_utils import run_method_and_decode
from .coregflux import CoRegFlux
from .prom import PROM
from .rfba import RFBA
from .srfba import SRFBA

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel

INTEGRATED_ANALYSIS_METHODS = {'rfba': RFBA,
                               'srfba': SRFBA}


def slim_rfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
              initial_state: Dict[str, float] = None,
              objective: Union[str, Dict[str, float]] = None,
              constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A Regulatory Flux Balance Analysis (RFBA) simulation of an integrated Metabolic-Regulatory model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value of
    the RFBA simulation.

    Fundamentals of the RFBA procedure:
        - A linear problem based on mass balance constraints (reactions/metabolites)
        - A synchronous simulation of the metabolic (GPRs) and regulatory networks is performed to obtain the
        metabolic state (boundaries of the reactions) and the regulatory state (boundaries of the regulators)
        - The metabolic state is used to constrain the reactions of the model
        - A FBA simulation is performed to obtain the fluxes of the model

    :param model: an integrated metabolic-regulatory model to be simulated
    :param initial_state: the initial state of the model. If None, the default initial state is used
    :param objective: the objective of the simulation. If None, the default objective is used
    :param constraints: the constraints of the model. If None, the default constraints are used
    :return: the objective value of the simulation
    """
    rfba = RFBA(model).build()

    objective_value, _ = run_method_and_decode(method=rfba, objective=objective, constraints=constraints,
                                               initial_state=initial_state)
    return objective_value


def slim_srfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
               initial_state: Dict[str, float] = None,
               objective: Union[str, Dict[str, float]] = None,
               constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A Synchronous Regulatory Flux Balance Analysis (SRFBA) simulation of an integrated Metabolic-Regulatory model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value of
    the SRFBA simulation.

    Fundamentals of the SRFBA procedure:
        - A linear problem based on reactions, GPRs, and regulatory interactions using mixed-integer constraints
        - Reactions, genes, and regulators are constrained by the bounds or coefficients
        - SRFBA is solved using a mixed-integer solver

    :param model: an integrated metabolic-regulatory model to be simulated
    :param initial_state: the initial state of the model. If None, the default initial state is used
    :param objective: the objective of the simulation. If None, the default objective is used
    :param constraints: the constraints of the model. If None, the default constraints are used
    :return: the objective value of the simulation
    """
    srfba = SRFBA(model).build()

    objective_value, _ = run_method_and_decode(method=srfba, objective=objective, constraints=constraints,
                                               initial_state=initial_state)
    return objective_value


def slim_prom(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
              initial_state: Dict[Tuple[str, str], float] = None,
              regulator: str = None,
              objective: Union[str, Dict[str, float]] = None,
              constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A Probabilistic Regulation of Metabolism (PROM) simulation of an integrated Metabolic-Regulatory model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value of
    the PROM simulation.

    Fundamentals of the PROM procedure:
        - A linear problem based on mass balance constraints (reactions/metabolites)
        - Reactions are further constrained using the regulator-target probabilities. The probabilities are
        obtained from the regulatory network. Namely, the probability of a gene being active when the its regulator
        is knocked out is used to constrain the associated reaction.
        - A FBA simulation is performed to obtain the flux distribution.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param initial_state: the initial state of the model. If None, the default initial state is used
    :param regulator: the regulator to be knocked out. If None, the first regulator is knocked out
    :param objective: the objective of the simulation. If None, the default objective is used
    :param constraints: the constraints of the model. If None, the default constraints are used
    :return: the objective value of the simulation
    """
    if not regulator:
        regulator = next(iter(model.regulators.keys()))

    prom = PROM(model).build()

    objective_value, _ = run_method_and_decode(method=prom, objective=objective, constraints=constraints,
                                               initial_state=initial_state, regulators=regulator)
    return objective_value


def slim_coregflux(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                   initial_state: Dict[str, float] = None,
                   objective: Union[str, Dict[str, float]] = None,
                   constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A CoRegFlux simulation of an integrated Metabolic-Regulatory model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value of
    the CoRegFlux simulation.

    Fundamentals of the CoRegFlux procedure:
         - A linear problem based on mass balance constraints (reactions/metabolites)
         - A linear regression estimator predicts the expression of target genes
         as function of the co-expression of regulators
         - the predicted expression of the target genes to constrain the bounds of the associated reactions
         - A FBA simulation is performed to obtain the flux distribution.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param initial_state: the initial state of the model. If None, the default initial state is used
    :param objective: the objective of the simulation. If None, the default objective is used
    :param constraints: the constraints of the model. If None, the default constraints are used
    :return: the objective value of the simulation
    """
    co_reg_flux = CoRegFlux(model).build()
    objective_value, _ = run_method_and_decode(method=co_reg_flux, objective=objective, constraints=constraints,
                                               initial_state=initial_state)
    return objective_value


def ifva(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
         fraction: float = 1.0,
         reactions: Sequence[str] = None,
         objective: Union[str, Dict[str, float]] = None,
         constraints: Dict[str, Tuple[float, float]] = None,
         initial_state: Dict[str, float] = None,
         method: str = 'srfba') -> pd.DataFrame:
    """
    Integrated Flux Variability Analysis (iFVA) of an integrated Metabolic-Regulatory model.
    iFVA is a flux variability analysis method that considers:
        - Metabolites
        - Reactions
        - Genes
        - Regulatory interactions
        - Target genes
        - Regulators
    It can be used to identify the reactions, genes, and regulators that are limiting the growth of a cell.
    In MEWpy, FVA is performed by solving a linear problem for each reaction in the model.
    The method can be either RFBA or SRFBA.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param fraction: the fraction of the optimal solution to be used as the upper bound for the objective function
    (default: 1.0)
    :param reactions: the reactions to be simulated. If None, all reactions are simulated (default: None)
    :param objective: the objective of the simulation. If None, the default objective is used (default: None)
    :param constraints: additional constraints to be added to the model. If None, no additional constraints are added
    :param method: the method to be used for the simulation. Available methods: 'rfba', 'srfba'. Default: 'srfba'
    :param initial_state: the initial state of the model. If None, the default initial state is used (default: None)
    :return: a DataFrame with the results of the simulation
    """
    if not reactions:
        reactions = model.reactions.keys()

    if objective:
        if hasattr(objective, 'keys'):
            obj = next(iter(objective.keys()))
        else:
            obj = str(objective)

    else:
        obj = next(iter(model.objective)).id

    if not constraints:
        constraints = {}

    LP = INTEGRATED_ANALYSIS_METHODS[method]

    _lp = LP(model).build()
    objective_value, _ = run_method_and_decode(method=_lp, objective=objective, constraints=constraints,
                                               initial_state=initial_state)
    constraints[obj] = (fraction * objective_value, ModelConstants.REACTION_UPPER_BOUND)

    lp = LP(model).build()

    result = defaultdict(list)
    for rxn in reactions:
        min_val, _ = run_method_and_decode(method=lp, objective={rxn: 1.0}, constraints=constraints, minimize=True)
        result[rxn].append(min_val)

        max_val, _ = run_method_and_decode(method=lp, objective={rxn: 1.0}, constraints=constraints, minimize=False)
        result[rxn].append(max_val)

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['minimum', 'maximum'])


def isingle_gene_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                          genes: Sequence[str] = None,
                          constraints: Dict[str, Tuple[float, float]] = None,
                          initial_state: Dict[str, float] = None,
                          method: str = 'srfba') -> pd.DataFrame:
    """
    Integrated single gene deletion analysis of an integrated Metabolic-Regulatory model.
    Integrated single gene deletion analysis is a method to determine the effect of deleting each gene
    in a model.
    It can be used to identify the genes that are essential for the growth of a cell.
    In MEWpy, single gene deletion analysis is performed by solving a linear problem for each gene in the model.
    A gene knockout can switch off reactions associated with the gene, only if the gene is essential for the reaction.
    The methods RFBA and SRFBA can determine if the gene is essential for the growth of a cell.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param genes: the genes to be simulated. If None, all genes are simulated (default: None)
    :param constraints: additional constraints to be added to the model. If None, no additional constraints are added
    :param initial_state: the initial state of the model. If None, the default initial state is used (default: None)
    :param method: the method to be used for the simulation. Available methods: 'rfba', 'srfba'. Default: 'srfba'
    :return: a DataFrame with the results of the simulation
    """
    if not initial_state:
        initial_state = {}

    if not genes:
        genes = model.genes.keys()

    LP = INTEGRATED_ANALYSIS_METHODS[method]
    lp = LP(model).build()

    result = {}
    for gene in genes:

        gene_coefficient = initial_state.pop(gene, None)
        initial_state[gene] = 0.0

        solution, status = run_method_and_decode(method=lp, constraints=constraints, initial_state=initial_state)

        result[gene] = [solution, status]

        if gene_coefficient is not None:
            initial_state[gene] = gene_coefficient
        else:
            initial_state.pop(gene)

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['growth', 'status'])


def isingle_reaction_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                              reactions: Sequence[str] = None,
                              constraints: Dict[str, Tuple[float, float]] = None,
                              initial_state: Dict[str, float] = None,
                              method: str = 'srfba') -> pd.DataFrame:
    """
    Integrated single reaction deletion analysis of an integrated Metabolic-Regulatory model.
    Integrated single reaction deletion analysis is a method to determine the effect of deleting each reaction
    in a model.
    It can be used to identify the reactions that are essential for the growth of a cell.
    In MEWpy, single reaction deletion analysis is performed by solving a linear problem for each reaction in the model.
    The methods RFBA and SRFBA can determine if the reaction is essential for the growth of a cell.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param reactions: the reactions to be simulated. If None, all reactions are simulated (default: None)
    :param constraints: additional constraints to be added to the model. If None, no additional constraints are added
    :param initial_state: the initial state of the model. If None, the default initial state is used (default: None)
    :param method: the method to be used for the simulation. Available methods: 'rfba', 'srfba'. Default: 'srfba'
    :return: a DataFrame with the results of the simulation
    """
    if not constraints:
        constraints = {}

    if not reactions:
        reactions = model.reactions.keys()

    LP = INTEGRATED_ANALYSIS_METHODS[method]
    lp = LP(model).build()

    result = {}
    for reaction in reactions:

        reaction_constraint = constraints.pop(reaction, None)
        constraints[reaction] = (0.0, 0.0)

        solution, status = run_method_and_decode(method=lp, constraints=constraints, initial_state=initial_state)

        result[reaction] = [solution, status]

        if reaction_constraint is not None:
            constraints[reaction] = reaction_constraint
        else:
            constraints.pop(reaction)

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['growth', 'status'])


def isingle_regulator_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                               regulators: Sequence[str] = None,
                               constraints: Dict[str, Tuple[float, float]] = None,
                               initial_state: Dict[str, float] = None,
                               method: str = 'srfba') -> pd.DataFrame:
    """
    Integrated single regulator deletion analysis of a regulatory model.
    Integrated single regulator deletion analysis is a method to determine the effect of deleting each regulator
    in a regulatory model.
    It can be used to identify the regulators that are essential for the growth of a cell.
    In MEWpy, single regulator deletion analysis is performed by solving a linear problem
    for each regulator in the model.
    A regulator knockout can switch off reactions associated with the regulator,
    only if the regulator is essential for the interactions and gprs.
    The methods RFBA and SRFBA can determine if the regulator is essential for the growth of a cell.

    :param model: an integrated metabolic-regulatory model to be simulated
    :param regulators: the regulators to be simulated. If None, all regulators are simulated (default: None)
    :param constraints: additional constraints to be added to the model. If None, no additional constraints are added
    :param initial_state: the initial state of the model. If None, the default initial state is used (default: None)
    :param method: the method to be used for the simulation. Available methods: 'rfba', 'srfba'. Default: 'srfba'
    :return: a DataFrame with the results of the simulation
    """
    if not initial_state:
        initial_state = {}

    if not regulators:
        regulators = model.regulators.keys()

    LP = INTEGRATED_ANALYSIS_METHODS[method]
    lp = LP(model).build()

    result = {}
    for regulator in regulators:

        regulator_coefficient = initial_state.pop(regulator, None)
        initial_state[regulator] = 0.0

        solution, status = run_method_and_decode(method=lp, constraints=constraints, initial_state=initial_state)

        result[regulator] = [solution, status]

        if regulator_coefficient is not None:
            initial_state[regulator] = regulator_coefficient
        else:
            initial_state.pop(regulator)

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['growth', 'status'])


def _decode_initial_state(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                          state: Dict[str, float]) -> Dict[str, float]:
    """
    Method responsible for retrieving the initial state of the model.
    The initial state is the state of all regulators found in the Metabolic-Regulatory model.
    :param model: the model to be simulated
    :param state: the initial state of the model
    :return: dict of regulatory/metabolic variable keys (regulators) and a value of 0 or 1
    """
    if not state:
        state = {}

    initial_state = {}
    for regulator in model.yield_regulators():
        if regulator.id in state:
            initial_state[regulator.id] = state[regulator.id]

        elif regulator.is_metabolite() and regulator.exchange_reaction:
            if regulator.exchange_reaction.id in state:
                initial_state[regulator.id] = state[regulator.exchange_reaction.id]

            else:
                initial_state[regulator.id] = abs(regulator.exchange_reaction.lower_bound)

        else:
            initial_state[regulator.id] = max(regulator.coefficients)

    return initial_state


def _decode_interactions(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                         state: Dict[str, float]) -> Dict[str, float]:
    """
    Decodes the state of the model to a dictionary with the state of each gene.
    :param model: the model to be decoded
    :param state: the state of the model
    :return: a dictionary with the state of each gene
    """
    target_state = {}
    for interaction in model.yield_interactions():

        for coefficient, event in interaction.regulatory_events.items():
            if event.is_none:
                continue

            result = event.evaluate(state)
            if result:
                target_state[interaction.target.id] = coefficient

            else:
                target_state[interaction.target.id] = 0.0

    return target_state


def _decode_gprs(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 state: Dict[str, float]) -> Dict[str, float]:
    """
    Decodes the state of the model to a dictionary with the state of each reaction.
    :param model: the model to be decoded
    :param state: the state of the model
    :return: a dictionary with the state of each gene
    """
    reaction_state = {}
    for reaction in model.yield_reactions():
        if reaction.gpr.is_none:
            continue

        result = reaction.gpr.evaluate(state)
        if result:
            reaction_state[reaction.id] = 1.0

        else:
            reaction_state[reaction.id] = 0.0

    return reaction_state


def find_conflicts(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                   strategy: str = 'two-step',
                   constraints: Dict[str, Tuple[float, float]] = None,
                   initial_state: Dict[str, float] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    It finds conflicts between the regulatory and metabolic states of an integrated GERM model.
    Setting up the regulators' initial state in integrated models is a difficult task. Most of the time,
    the initial state is not known and hinders feasible solutions during simulation.
    This method can be used to find regulatory states that affect the growth of the cell.

    One can use two strategies to find conflicts:
        - One-step strategy: it simulates the model with the initial state of the regulators and
        checks if the metabolic state violates essential genes and reactions.
        - Two-step strategy: it simulates the model with the initial state of the regulators to retrieve the regulatory
        state of all regulators. Then, it simulates the model with the regulatory state of the regulators to retrieve
        the metabolic state. Finally, it checks if the metabolic state violates essential genes and reactions.

    NOTE: Although this method finds direct conflicts between regulators and genes affecting cellular growth,
    it might not detect indirect conflicts.
    For example, an active regulator can repress an essential metabolic gene and thus being reported as a conflict.
    However, this regulator can only be active due to an erroneous initial state of another regulator or environmental
    stimuli, metabolite or reaction. That is, this active regulator is also a target being affected by other regulators.
    In these cases, these regulators should be reported as conflicts as well.
    Hence, this method should be run several times with different initial states
    of the regulators to find all conflicts!!!

    :param model: a metabolic-regulatory model to be simulated
    :param strategy: the strategy to be used to find conflicts. Available strategies: 'one-step', 'two-step'.
    one-step performs a single regulatory simulation. two-step performs two simulations. Default: 'two-step'
    :param constraints: additional constraints to be added to the model. If None, no additional constraints are added
    :param initial_state: the initial state of the model. If None, the default initial state is used (default: None)
    :return: a DataFrame with the conflicts between essential genes and regulatory states, and a DataFrame with the
    conflicts between essential reactions and metabolic states
    """
    if not constraints:
        constraints = {}
    else:
        constraints = constraints.copy()

    if not initial_state:
        initial_state = {}
    else:
        initial_state = initial_state.copy()

    # 1. it performs a FBA simulation to find the optimal growth rate
    from mewpy.germ.analysis import FBA
    solution = FBA(model).build().optimize(solver_kwargs={'constraints': constraints})

    if not solution.objective_value:
        raise RuntimeError('FBA solution is not feasible (objective value is 0). To find inconsistencies, '
                           'the metabolic model must be feasible.')

    # 2. it performs an essential genes analysis using FBA
    from mewpy.germ.analysis.metabolic_analysis import single_gene_deletion
    gene_deletion = single_gene_deletion(model, constraints=constraints)
    essential_genes = gene_deletion[gene_deletion['growth'] < ModelConstants.TOLERANCE]

    from mewpy.germ.analysis.metabolic_analysis import single_reaction_deletion
    reaction_deletion = single_reaction_deletion(model, constraints=constraints)
    essential_reactions = reaction_deletion[reaction_deletion['growth'] < ModelConstants.TOLERANCE]

    state = _decode_initial_state(model, initial_state)

    # noinspection PyTypeChecker
    regulatory_state = _decode_interactions(model, state)

    if strategy == 'two-step':
        regulatory_state = {gene: value for gene, value in regulatory_state.items() if model.get(gene).is_regulator()}
        regulatory_state = {**state, **regulatory_state}
        # noinspection PyTypeChecker
        metabolic_state = _decode_interactions(model, regulatory_state)

    else:
        metabolic_state = regulatory_state.copy()

    deleted_genes = set(gene for gene, val in metabolic_state.items() if val < ModelConstants.TOLERANCE)
    gene_conflicts = set.intersection(deleted_genes, set(essential_genes.index))

    reaction_state = _decode_gprs(model, metabolic_state)
    deleted_reactions = set(reaction for reaction, val in reaction_state.items() if val < ModelConstants.TOLERANCE)
    reaction_conflicts = set.intersection(deleted_reactions, set(essential_reactions.index))

    repressed_genes = []
    for gene in gene_conflicts:
        gene = model.genes[gene]

        if gene.is_target():
            # noinspection PyUnresolvedReferences
            repressed_gene = {regulator: regulatory_state[regulator] for regulator in gene.regulators}
            # noinspection PyUnresolvedReferences
            repressed_gene['interaction'] = str(gene.interaction)
        else:
            repressed_gene = {}

        df = pd.DataFrame(repressed_gene, index=[gene.id])
        repressed_genes.append(df)

    if repressed_genes:
        repressed_genes = pd.concat(repressed_genes)
        cols = ['interaction'] + [col for col in repressed_genes.columns if col != 'interaction']
        repressed_genes = repressed_genes[cols]
    else:
        repressed_genes = pd.DataFrame()

    repressed_reactions = []
    for reaction in reaction_conflicts:
        reaction = model.reactions[reaction]

        repressed_reaction = {gene: metabolic_state[gene] for gene in reaction.genes}
        repressed_reaction['gpr'] = reaction.gene_protein_reaction_rule

        df = pd.DataFrame(repressed_reaction, index=[reaction.id])
        repressed_reactions.append(df)

    if repressed_reactions:
        repressed_reactions = pd.concat(repressed_reactions)
        cols = ['gpr'] + [col for col in repressed_reactions.columns if col != 'gpr']
        repressed_reactions = repressed_reactions[cols]
    else:
        repressed_reactions = pd.DataFrame()

    return repressed_genes, repressed_reactions



from typing import Union

from mewpy.solvers import get_default_solver
from mewpy.solvers.sglobal import __MEWPY_solvers__ as solvers
from mewpy.solvers.solver import Solver


integer_coefficients = ((0, 0), (1, 1), (0.0, 0.0), (1.0, 1.0), (0, 1), (0.0, 1.0))


def get_solver_instance(solver: Union[str, Solver] = None) -> Solver:
    """
    It returns a new empty mewpy solver instance. However, if a solver instance is provided,
    it only checks if it is a mewpy solver.
    :param solver: Solver, CplexSolver, GurobiSolver or OptLangSolver instance or name of the solver
    :return: a mewpy solver instance
    """
    if solver is None:
        solver_name = get_default_solver()

        SolverType = solvers[solver_name]

        solver = SolverType()

    elif isinstance(solver, str):

        SolverType = solvers.get(solver, None)

        if SolverType is None:
            raise ValueError(f'{solver} is not listed as valid solver. Check the valid solvers: {solvers}')

        solver = SolverType()

    elif isinstance(solver, Solver):

        pass

    else:
        raise ValueError(f'Invalid solver {solver}. Check the valid solvers: {solvers}')

    return solver


class Node:

    def __init__(self, value, length=None, idxes=None):

        if not length:
            length = 0

        if not idxes:
            idxes = None

        self._next = None
        self._previous = None

        self.value = value
        self.length = length
        self.idxes: slice = idxes

    def __str__(self):
        return self.value

    @property
    def next(self):
        return self._next

    @property
    def previous(self):
        return self._previous

    def unlink(self):
        self._next = None
        self._previous = None


class LinkedList:

    def __init__(self, *args):

        if args:
            head = args[0]
            tail = args[-1]
            nodes = list(args)
            nodes.append(None)
            nodes = list(zip(nodes[:-1], nodes[1:]))

        else:
            head = None
            tail = None
            nodes = []

        self._data = {}

        self._head = head
        self._tail = tail

        for node, next_node in nodes:
            node._next = next_node
            if next_node:
                next_node._previous = node

        self.build_data()

    @property
    def data(self):
        return self._data

    def __len__(self):

        res = 0

        if self._tail:

            res = self._data.get(self._tail.value).idxes.stop

        elif self._head:

            res = self._data.get(self._tail.value).idxes.stop

        if res > 0:
            return res

        return 0

    def __hash__(self):
        return self._data.__hash__()

    def __eq__(self, other):
        return self._data.__eq__(other)

    def __contains__(self, item):
        return self._data.__contains__(item)

    def __getitem__(self, item):

        return self._data.__getitem__(item).idxes

    def __setitem__(self, key, value):

        raise NotImplementedError('Linked lists do not support item setting. Try pop or add')

    def get(self, value, default=None):

        node = self._data.get(value, None)

        if node:
            return node.idxes

        return default

    def keys(self, unique=True):

        if unique:
            yield from self._data.keys()

            return

        for key, node in self._data.items():

            if node.idxes.stop - node.idxes.start > 1:

                for i in range(node.idxes.start, node.idxes.stop):
                    yield f'{key}_{i}'

            else:
                yield key

    def values(self):

        return (node.idxes for node in self._data.values())

    def items(self):

        return ((key, node.idxes) for key, node in self._data.items())

    def traverse(self):

        node = self._head

        while node is not None:
            yield node
            node = node.next

    def map(self, function):

        node = self._head

        while node is not None:
            function(node)
            node = node.next

    def get_node(self, value, default=None):

        return self._data.get(value, default)

    def build_data(self):

        self._data = {}

        node = self._head

        start = 0
        while node is not None:
            stop = start + node.length

            node.idxes = slice(start, stop)

            self._data[node.value] = node

            start = stop

            node = node.next

    def extend(self, nodes):

        for node in nodes:
            self.add(node)

    def add(self, node):

        if isinstance(node, (tuple, list)):
            node = Node(node[0], node[1])

        elif isinstance(node, dict):
            node = Node(node['value'], node['length'])

        elif isinstance(node, Node):
            pass

        else:
            raise TypeError('Node must be a tuple, list, dict(value=val, length=len) or Node instance')

        if node.value in self.data:
            raise ValueError('Node value is already in linked list')

        if not self._head:

            node._previous = None
            node._next = None

            self._head = node
            self._tail = node

            if not node.idxes:
                node.idxes = slice(0, node.length)

            self.data[node.value] = node

        else:

            if not node.idxes:
                # noinspection PyProtectedMember
                node.idxes = slice(self._tail.idxes.stop, self._tail.idxes.stop + node.length)

            self.data[node.value] = node

            node._previous = self._tail
            node._next = None

            self._tail._next = node
            self._tail = node

    def pop(self, value):

        if isinstance(value, Node):
            # noinspection PyUnresolvedReferences
            value = Node.value

        node = self._data.pop(value)
        previous_node = node.previous
        next_node = node.next

        if previous_node and next_node:

            previous_node._next = next_node
            next_node._previous = previous_node

            node._next = None
            node._previous = None

            start = previous_node.idxes.stop

        elif previous_node and not next_node:

            previous_node._next = None

            node._next = None
            node._previous = None

            self._tail = previous_node

            return node

        elif not previous_node and next_node:

            next_node._previous = None

            node._next = None
            node._previous = None

            self._head = next_node

            start = 0

        else:

            node._next = None
            node._previous = None

            self._head = None
            self._tail = None

            self._data = {}

            return node

        _node = next_node

        while _node is not None:
            stop = start + _node.length

            _node.idxes = slice(start, stop)

            self._data[_node.value] = _node

            start = stop

            _node = _node.next

        return node

    def clear(self):

        self._data = {}

        self.map(lambda n: n.unlink())

        self._tail = None
        self._head = None



from jmetal.algorithm.multiobjective import NSGAII, SPEA2
from jmetal.algorithm.multiobjective.nsgaiii import NSGAIII
from jmetal.algorithm.multiobjective.nsgaiii import UniformReferenceDirectionFactory
from jmetal.algorithm.singleobjective import GeneticAlgorithm, SimulatedAnnealing
from jmetal.operator import BinaryTournamentSelection
from jmetal.util.termination_criterion import StoppingByEvaluations

from .observers import PrintObjectivesStatObserver, VisualizerObserver
from .problem import JMetalKOProblem, JMetalOUProblem
from .settings import get_population_size
from ..ea import AbstractEA, Solution
from mewpy.util.constants import EAConstants
from mewpy.util.process import get_evaluator, cpu_count
import numpy as np

# SOEA alternatives
soea_map = {
    'GA': GeneticAlgorithm,
    'SA': SimulatedAnnealing
}
# MOEA alternatives
moea_map = {
    'NSGAII': NSGAII,
    'SPEA2': SPEA2,
    'NSGAIII': NSGAIII
}


class EA(AbstractEA):
    """
    EA running helper for JMetal.

    :param problem: The optimization problem.
    :param initial_population: (list) The EA initial population.
    :param max_generations: (int) The number of iterations of the EA (stopping criteria).
    """

    def __init__(self, problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS, mp=True,
                 visualizer=False, algorithm=None, **kwargs):

        super(EA, self).__init__(problem, initial_population=initial_population,
                                 max_generations=max_generations, mp=mp, visualizer=visualizer, **kwargs)

        self.algorithm_name = algorithm

        if self.problem.strategy == 'KO':
            self.ea_problem = JMetalKOProblem(self.problem, self.initial_population)
        else:
            self.ea_problem = JMetalOUProblem(self.problem, self.initial_population)
        self.crossover, self.mutation = self.ea_problem.build_operators()
        self.population_size = kwargs.get('population_size', get_population_size())
        self.max_evaluations = self.max_generations * self.population_size

        s = []
        for f in self.problem.fevaluation:
            if f.maximize:
                s.append(-1)
            else:
                s.append(1)
        self._sense = np.array(s)


    def get_population_size(self):
        return self.population_size

    def _run_so(self):
        """ Runs a single objective EA optimization ()
        """
        self.ea_problem.reset_initial_population_counter()
        if self.algorithm_name == 'SA':
            print("Running SA")
            self.mutation.probability = 1.0
            algorithm = SimulatedAnnealing(
                problem=self.ea_problem,
                mutation=self.mutation.probability,
                termination_criterion=StoppingByEvaluations(max_evaluations=self.max_evaluations)
            )

        else:
            print("Running GA")
            algorithm = GeneticAlgorithm(
                problem=self.ea_problem,
                population_size=self.population_size,
                offspring_population_size=self.population_size,
                mutation=self.mutation,
                crossover=self.crossover,
                selection=BinaryTournamentSelection(),
                termination_criterion=StoppingByEvaluations(
                    max_evaluations=self.max_evaluations)
            )

        algorithm.observable.register(observer=PrintObjectivesStatObserver())
        self.algorithm = algorithm
        algorithm.run()

        result = algorithm.solutions
        return result

    def _run_mo(self):
        """ Runs a multi objective EA optimization
        """
        self.ea_problem.reset_initial_population_counter()
        if self.algorithm_name in moea_map.keys():
            f = moea_map[self.algorithm_name]
        else:
            if self.ea_problem.number_of_objectives > 2:
                self.algorithm_name == 'NSGAIII'
            else:
                f = moea_map['SPEA2']

        args = {
            'problem': self.ea_problem,
            'population_size': self.population_size,
            'mutation': self.mutation,
            'crossover': self.crossover,
            'termination_criterion': StoppingByEvaluations(max_evaluations=self.max_evaluations)
        }

        if self.mp:
            args['population_evaluator'] = get_evaluator(self.ea_problem, n_mp=cpu_count())

        print(f"Running {self.algorithm_name}")
        if self.algorithm_name == 'NSGAIII':
            args['reference_directions'] = UniformReferenceDirectionFactory(self.ea_problem.number_of_objectives,
                                                                            n_points=self.population_size-1)
            algorithm = NSGAIII(**args)
        else:
            args['offspring_population_size'] = self.population_size
            algorithm = f(**args)

        if self.visualizer:
            axis_labels = [f.short_str() for f in self.problem.fevaluation]
            algorithm.observable.register(observer=VisualizerObserver(axis_labels=axis_labels))
        else:
            algorithm.observable.register(observer=PrintObjectivesStatObserver())
        self.algorithm = algorithm
        algorithm.run()
        result = algorithm.solutions
        return result

    def _correct_sense(self, fitness):
        return list(np.array(fitness)*self._sense)

    def _convertPopulation(self, population):
        """Converts a population represented in Inpyred format to
        MEWpy solution format.

        :param list population: A list of solutions.

        :returns: A MEWpy list of solutions.
        """
        p = []
        for i in range(len(population)):
            # Corrects fitness values for maximization problems            
            obj = self._correct_sense([x for x in population[i].objectives])

            val = set(population[i].variables[:])
            values = self.problem.decode(val)
            const = self.problem.solution_to_constraints(values)
            solution = Solution(values, obj, const)
            p.append(solution)
        return p

    def _get_current_population(self):
        """Dumps the population for gracefull exit."""
        pop = self.algorithm.solutions
        cv = self._convertPopulation(pop)
        return cv



"""
Implements a crossmodel simulation of solutions
"""
import pandas as pd


class NotationTranslator:
    def __init__(self, database, from_notation, to_notation, admissible=None, sep=';'):
        if isinstance(database, pd.DataFrame):
            self.db = database
        elif isinstance(database, str):
            self.db = pd.read_csv(database, sep=sep)
        self.from_notation = from_notation
        self.to_notation = to_notation
        self.admissible = admissible

    def columns(self):
        return self.db.columns.tolist()

    def translate(self, value):
        """
        Translate a single value
        """

        la = self.db.loc[self.db[self.from_notation].str.contains(value, na=False)][self.from_notation].tolist()
        lb = None
        if la and len(la) > 0:
            if len(la) == 1:
                lb = self.db.loc[self.db[self.from_notation] == la[0]][self.to_notation].tolist()
            else:
                for x in la:
                    tokens = x.split(' ')
                    if value in tokens:
                        lb = self.db.loc[self.db[self.from_notation] == x][self.to_notation].tolist()
                        break
            if not lb:
                raise ValueError(f'Value {value} not found.')
            if len(lb) > 1:
                raise ValueError(f'More than a value {value} found. {lb}')
            s = lb[0]
            tokens = s.split(' ')
            if tokens and len(tokens) == 1:
                return tokens[0]
            else:
                idx = 0
                while idx < len(tokens):
                    res = tokens[idx]
                    if res in self.admissible:
                        return res
                    idx += 1
                raise ValueError(f'Value {value} correspondences not in admissible.')
        else:
            raise ValueError(f'Value {value} not found.')

    def translate_representation(self, representation, source_prefix="", destination_prefix=""):
        """
        Translates constraints.
        constraints are defined as a dictionary (OU) or a list (KO)
        """

        p = len(source_prefix)
        if isinstance(representation, list):
            return [destination_prefix + self.translate(value[p:]) for value in representation]
        else:
            return {destination_prefix + self.translate(value[p:]): level for value, level in representation.items()}

    def get_list(self, name):
        return self.db[name].tolist()



import math
import warnings
from abc import ABCMeta, abstractmethod
from functools import reduce
import numpy as np

from mewpy.solvers.ode import ODEStatus
from ..simulation import get_simulator, SimulationMethod, SStatus
from ..util.constants import EAConstants, ModelConstants


class EvaluationFunction:
    __metaclass__ = ABCMeta

    def __init__(self, maximize=True, worst_fitness=0.0):
        """This abstract class should be extended by all evaluation functions.
        """
        self.worst_fitness = worst_fitness
        self.maximize = maximize

    @abstractmethod
    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """
        raise NotImplementedError

    @abstractmethod
    def method_str(self):
        raise NotImplementedError

    def short_str(self):
        return self.method_str

    def __str__(self):
        return self.method_str()

    @abstractmethod
    def required_simulations(self):
        return None

    @property
    def no_solution(self):
        """
        Value to be retuned for wost case evaluation
        """
        if self.worst_fitness is not None:
            res = self.worst_fitness
        elif self.maximize:
            res = -math.inf
        else:
            res = math.inf
        return res

    def __call__(self, simulationResult, candidate, **kwargs):
        return self.get_fitness(simulationResult, candidate, **kwargs)


class PhenotypeEvaluationFunction(EvaluationFunction):

    def __init__(self, maximize=True, worst_fitness=0.0):
        super(PhenotypeEvaluationFunction, self).__init__(maximize=maximize, worst_fitness=worst_fitness)


class KineticEvaluationFunction(EvaluationFunction):

    def __init__(self, maximize=True, worst_fitness=0.0):
        super(KineticEvaluationFunction, self).__init__(maximize=maximize, worst_fitness=worst_fitness)


class TargetFlux(PhenotypeEvaluationFunction,KineticEvaluationFunction):
    """ Target Flux evaluation function.
    The fitness value is the flux value of the identified reaction.
    If the reaction parameter is None, the fitness value is the optimization objective value.
    Additional parameters include a minimum of allowed biomass value computed from the min_biomass_per
    and reference flux values

    :param reaction: (str) The reaction identifier whose flux value is to be used as fitness. Default None \
        in which case the model objective is considered.
    :param biomass: (str) The biomass reaction identifier.
    :param maximize: (boolean) The optimization direction. Default True for maximization.
    :param min_biomass_value: (float) The minimum biomass value.
    :param min_biomass_per: (float) Minimum biomass percentage. Only used if no min_biomass_value is provided.

    """

    def __init__(self, reaction, biomass=None, maximize=True, min_biomass_value=None, min_biomass_per=0.0,
                 method=SimulationMethod.pFBA):
        super(TargetFlux, self).__init__(maximize=maximize, worst_fitness=0.0)
        self.reaction = reaction
        self.biomass = biomass
        self.min_biomass_value = min_biomass_value
        self.min_biomass_per = min_biomass_per
        self.method = method
        self.kinetic = False

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """
        if self.kinetic:
            sim = simul_results
        else:
            sim = simul_results[self.method] if self.method in simul_results.keys(
            ) else None

        if not sim or sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL, ODEStatus.OPTIMAL):
            return self.no_solution

        # only executed once if required
        if self.biomass and self.min_biomass_value is None:
            self.min_biomass_value = 0.0
            if self.min_biomass_per > 0.0:
                simulation = get_simulator(
                    sim.model, envcond=sim.envcond, constraints=sim.model_constraints)
                result = simulation.simulate(objective={self.biomass: 1})
                self.min_biomass_value = self.min_biomass_per * \
                    result.fluxes[self.biomass]

        if self.biomass and sim.fluxes[self.biomass] < self.min_biomass_value:
            res = self.no_solution
        elif self.reaction and self.reaction in sim.fluxes.keys():
            res = sim.fluxes[self.reaction]
        else:
            res = sim.objective_value

        return res

    def required_simulations(self):
        """
        If the evaluation function requires a pre-simulation to compute fitness values
        """
        return [self.method]

    def short_str(self):
        return "TargetFlux"

    def method_str(self):
        return "TargetFlux {} with at least {} of biomass ({})".format(self.reaction, self.min_biomass_per,
                                                                       self.biomass)


class WYIELD(PhenotypeEvaluationFunction):
    """ Weighted Yield (WYIELD) objective function, a linear combination of the target product minimum and
    maximum FVA under the introduced metabolic modifications.

    :param biomassId: (str) Biomass reaction identifier.
    :param productId: (str) Target product reaction identifier.

    kwargs options:

    :param min_biomass_value: (float) Minimum biomass value (default None, in which case the min_biomass_per is used).
    :param min_biomass_per: (float) Instead of defining explicitly a minimum biomass value, a percentage of the wild \
        type biomass is used. Only used when no min_biomass_value is defined (default min_biomass_per: 0.10).
    :param alpha: (float) Tradeoff between the Max and min FVA of the target product (alpha Max + (1-alpha) min). \
        Must be in range [0,1]  (default alpha: 0.3).
    :param scale: (boolean) Defines if the WYIELD is devided by the biomass of the simulated result (default false).

    """

    def __init__(self, biomassId, productId, maximize=True, **kwargs):
        super(WYIELD, self).__init__(maximize=maximize, worst_fitness=0.0)
        self.biomassId = biomassId
        self.productId = productId
        # parameters
        self.min_biomass_value = kwargs.get('min_biomass_value', None)
        self.min_biomass_per = kwargs.get('min_biomass_per', 0.1)
        self.scale = kwargs.get('scale', False)
        self.method = SimulationMethod.FBA
        self.alpha = kwargs.get('alpha', 0.3)
        self.obj_frac = kwargs.get('obj_frac', 0.99)
        if self.alpha > 1 or self.alpha < 0:
            warnings.warn(
                "The value of the tradeoff parameter alpha should be in range 0 to 1. Setting default value.")
            self.alpha = 0.3

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """

        sim = simul_results[self.method] if self.method in simul_results.keys(
        ) else None
        if not sim or sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
            return self.no_solution

        fvaMaxProd = 0.0
        fvaMinProd = 0.0
        scalefactor = kwargs.get('scalefactor', None)

        model = sim.model
        ssFluxes = sim.fluxes

        simulation = get_simulator(
            model, envcond=sim.envcond, constraints=sim.model_constraints)

        if not ssFluxes:
            return self.no_solution
        ids = list(ssFluxes.keys())
        if self.biomassId not in ids or self.productId not in ids:
            raise ValueError(
                "Reaction ids are not present in the fluxes distribution.")

        biomassFluxValue = ssFluxes[self.biomassId] * self.obj_frac

        try:
            # computed only once
            if self.min_biomass_value is None or self.min_biomass_value < 0.0:
                solution = simulation.simulate(
                    objective={self.biomassId: 1}, scalefactor=scalefactor)
                wtBiomassValue = solution.fluxes[self.biomassId]
                minBiomass = wtBiomassValue * self.min_biomass_per
                self.min_biomass_value = minBiomass
            else:
                minBiomass = self.min_biomass_value

            constraints = {}
            constraints.update(sim.simulation_constraints)
            # add biomass constraint
            constraints[self.biomassId] = (biomassFluxValue, ModelConstants.REACTION_UPPER_BOUND)

            # only need to simulate FVA max if alpha is larger than 0, otherwise it will always be zero
            if (self.alpha > 0):
                fvaMaxResult = simulation.simulate(
                    objective={self.productId: 1}, constraints=constraints, scalefactor=scalefactor)
                if fvaMaxResult.status == SStatus.OPTIMAL:
                    fvaMaxProd = fvaMaxResult.fluxes[self.productId]
                else:
                    return self.no_solution

            # only need to simulate FVA min if alpha is lesser than 1, otherwise it will always be zero
            if (self.alpha < 1):
                fvaMinResult = simulation.simulate(objective={
                    self.productId: 1}, constraints=constraints, maximize=False, scalefactor=scalefactor)
                if fvaMinResult.status == SStatus.OPTIMAL:
                    fvaMinProd = fvaMinResult.fluxes[self.productId]
                else:
                    return self.no_solution

            res = self.no_solution
            if EAConstants.DEBUG:
                print(f"WYIELD FVA max: {fvaMaxProd} min:{fvaMinProd}")
            if biomassFluxValue > minBiomass:
                res = (self.alpha * fvaMaxProd + (1 - self.alpha) * fvaMinProd)
                if self.scale:
                    res = res / biomassFluxValue
            return res
        except Exception:
            return self.no_solution

    def required_simulations(self):
        return [self.method]

    def short_str(self):
        return "WYIELD"

    def method_str(self):
        return "WYIELD biomass: {} product: {} ".format(self.biomassId, self.productId)


class BPCY(PhenotypeEvaluationFunction):
    """
    This class implements the "Biomass-Product Coupled Yield" objective function. The fitness is given by the equation:
    (biomass_flux * product_flux)/ uptake_flux

    :param biomass: (str) Biomass reaction identifier
    :param product: (str) Target product reaction identifier
    :param uptake: (str) (optional) Reaction of uptake. If no substract is defined, ie uptake is None, a substract \
        flux value of 1.0 is considered.

    kargs options:

    :param method: (SimulationMethod) The simulation method. Default Node in which case received simulation results \
        are used to compute the biomass product coupled yield.
    :param reference: (dic) Wild type reference values when MOMA, lMOMA or ROOM are defined as method. \
        If not provided, wild type reference values will be computed.

    """

    def __init__(self, biomass, product, uptake=None, maximize=True, **kwargs):
        super(BPCY, self).__init__(maximize=maximize, worst_fitness=0.0)
        self.biomassId = biomass
        self.productId = product
        self.uptakeId = uptake
        self.method = kwargs.get('method', SimulationMethod.pFBA)
        self.reference = kwargs.get('reference', None)
        self.worst_fitness = 0.0

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """

        sim = simul_results[self.method] if self.method in simul_results.keys(
        ) else None
        if not sim or sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
            return self.no_solution

        ssFluxes = sim.fluxes

        ids = list(ssFluxes.keys())
        if self.biomassId not in ids or self.productId not in ids:
            raise ValueError(
                "Biomass or product reactions ids are not present in the fluxes distribution.")

        if self.uptakeId and self.uptakeId in ids:
            uptake = abs(ssFluxes[self.uptakeId])
        else:
            uptake = 1.0

        if uptake == 0.0:
            return self.no_solution
        if EAConstants.DEBUG:
            try:
                print("BPCY Bionamss: {} product: {}".format(ssFluxes[self.biomassId], ssFluxes[self.productId]))
            except Exception:
                print("BPCY No Fluxes")
        return (ssFluxes[self.biomassId] * ssFluxes[self.productId]) / uptake

    def required_simulations(self):
        return [self.method]

    def short_str(self):
        return "BPCY"

    def method_str(self):
        if self.uptakeId:
            return "BPCY (" + self.biomassId + " * " + self.productId + ") / " + self.uptakeId
        else:
            return "BPCY (" + self.biomassId + " * " + self.productId + ")"


class BPCY_FVA(PhenotypeEvaluationFunction):
    """
    This class implements the "Biomass-Product Coupled Yield" objective function with FVA as defined in
    "OptRAM: In-silico strain design via integrative regulatory-metabolic network modeling".
    It combines BPCY with WYIELD objective functions.

    The fitness is given by the equation:
    ((biomass_flux * product_flux) * / uptake_flux ) * (1-log((range)/(target))) where range=(FVA_max-FVA_min)/2
    and target= (FVA_max+FVA_min)/2

    :param biomass: (str) Biomass reaction identifier.
    :param product: (str) Target product reaction identifier.
    :param uptake: (str) (optional) Reaction of uptake. If no substract is defined, ie uptakeId is None, a substract \
        flux value of 1.0 is considered.

    kwargs options:

    :param method: (SimulationMethod) The simulation method. Default Node in which case received simulation results \
        are used to compute the biomass product coupled yield.
    :param reference: (dic) Wild type reference values when MOMA, lMOMA or ROOM are defined as method. \
        If not provided, wild type reference values will be computed.

    """

    def __init__(self, biomass, product, uptake=None, maximize=True, **kwargs):
        super(BPCY_FVA, self).__init__(maximize=maximize, worst_fitness=0.0)
        self.biomassId = biomass
        self.productId = product
        self.uptakeId = uptake
        self.method = kwargs.get('method', SimulationMethod.pFBA)
        self.reference = kwargs.get('reference', None)
        self.worst_fitness = 0.0
        self.obj_frac = kwargs.get('obj_frac', 0.99)

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate.

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """
        sim = simul_results[self.method] if self.method in simul_results.keys(
        ) else None
        if not sim or sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
            return self.no_solution

        ssFluxes = sim.fluxes

        ids = list(ssFluxes.keys())
        if self.biomassId not in ids or self.productId not in ids:
            raise ValueError(
                "Biomass or product reactions ids are not present in the fluxes distribution.")

        if self.uptakeId and self.uptakeId in ids:
            uptake = abs(ssFluxes[self.uptakeId])
        else:
            uptake = 1.0

        if uptake == 0.0:
            return self.no_solution

        # computes target FVA min and max
        simulation = get_simulator(
            sim.model, envcond=sim.envcond, constraints=sim.model_constraints)
        v_min = 0
        v_max = 1
        constraints = sim.simulation_constraints
        # add biomass constraint
        biomassFluxValue = ssFluxes[self.biomassId] * self.obj_frac
        constraints[self.biomassId] = (biomassFluxValue, ModelConstants.REACTION_UPPER_BOUND)

        fvaMaxResult = simulation.simulate(
            objective={self.productId: 1}, constraints=constraints)
        v_max = fvaMaxResult.fluxes[self.productId]

        if not v_max:
            return self.worst_fitness

        fvaMinResult = simulation.simulate(
            objective={self.productId: 1}, constraints=constraints, maximize=False)
        v_min = fvaMinResult.fluxes[self.productId]

        if abs(v_max) == abs(v_min):
            return (ssFluxes[self.biomassId] * ssFluxes[self.productId]) / uptake
        else:
            return ((ssFluxes[self.biomassId] * ssFluxes[self.productId]) / uptake) * (
                1 - math.log(abs((v_max - v_min) / (v_max + v_min))))

    def required_simulations(self):
        return [self.method]

    def short_str(self):
        return "BPCY_FVA"

    def method_str(self):
        if self.uptakeId:
            return "BPCY_FVA (" + self.biomassId + " * " + self.productId + ") / " + self.uptakeId
        else:
            return "BPCY_FVA (" + self.biomassId + " * " + self.productId + ")"


class AggregatedSum(PhenotypeEvaluationFunction, KineticEvaluationFunction):
    """
    Aggredated sum evaluation function. Used to converte MOEAs into Single Objective EAs.

    :param fevaluation: (list) List of evaluation functions.
    :param tradeoffs: (list) Tradeoff values for each evaluation function. If None, all functions have \
        the same associated weight.

    """

    def __init__(self, fevaluation, tradeoffs=None, maximize=True):
        super(AggregatedSum, self).__init__(
            maximize=maximize, worst_fitness=0.0)
        self.fevaluation = fevaluation
        if tradeoffs and len(tradeoffs) == len(fevaluation):
            self.tradeoffs = tradeoffs
        else:
            self.tradeoffs = [1 / len(self.fevaluation)] * \
                (len(self.fevaluation))

    def required_simulations(self):
        methods = []
        for f in self.fevaluation:
            methods.extend(f.required_simulations())
        return list(set(methods))

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """
        res = []
        for f in self.fevaluation:
            res.append(f.get_fitness(simul_results, candidate, **kwargs))
        # return sum(map(lambda x, y: x * y, f, self.tradeoffs))
        return np.dot(res, self.tradeoffs)

    def short_str(self):
        return "Agg"

    def method_str(self):
        return "Aggregated Sum = " + reduce(lambda a, b: a + " " + b, [f.method_str() for f in self.fevaluation], "")


class CandidateSize(PhenotypeEvaluationFunction, KineticEvaluationFunction):
    """
    Maximize/minimize the number of modifications (reactions, genes, enzymes, etc.).

    :param (bool) maximize: Optimization sense. Default False (minimize)

    """

    def __init__(self, maximize=False):
        super(CandidateSize, self).__init__(maximize=maximize, worst_fitness=0.0)

    def get_fitness(self, simulResult, candidate, **kwargs):
        return len(candidate)

    def required_simulations(self):
        """
        If the evaluation function requires a pre-simulation to compute fitness values
        """
        return []

    def short_str(self):
        return "Size"

    def method_str(self):
        return "Minimize/maximize the number of alterations"


class MinCandSize(CandidateSize):

    def __init__(self, maximize=False):
        warnings.warn("This class will soon be depricated. Use CandidateSize instead.")
        super(MinCandSize, self).__init__(maximize=maximize, worst_fitness=0.0)


class CNRFA(PhenotypeEvaluationFunction):
    """Counts the Number of Reaction Fluxes Above a specified value.

    :param PhenotypeEvaluationFunction: [description]
    :type PhenotypeEvaluationFunction: [type]
    """

    def __init__(self, reactions, threshold=0.1, maximize=True, **kwargs):
        super(CNRFA, self).__init__(maximize=maximize, worst_fitness=0)
        self.reactions = reactions
        self.method = kwargs.get('method', SimulationMethod.pFBA)
        self.theshold = threshold

    def get_fitness(self, simul_results, candidate, **kwargs):
        """Evaluates a candidate

        :param simul_results: (dic) A dictionary of phenotype SimulationResult objects
        :param candidate:  Candidate beeing evaluated
        :returns: A fitness value.

        """
        sim = simul_results[self.method] if self.method in simul_results.keys(
        ) else None
        if not sim or sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
            return self.no_solution

        count = 0
        for rxn in self.reactions:
            if sim.fluxes[rxn]> self.theshold:
                count +=1
        return count

    def required_simulations(self):
        return [self.method]

    def short_str(self):
        return "CNRFA"

    def method_str(self):
        return "Count N Reaction Fluxes Above"
    

class ModificationType(PhenotypeEvaluationFunction, KineticEvaluationFunction):
    """This Objective function favors solutions with deletions, under expression and over expression,
    in this same order.
    """

    def __init__(self, penalizations={'KO': 5, 'UE': 2, 'OE': 0}, maximize=True):
        super(ModificationType, self).__init__(maximize=maximize, worst_fitness=0.0)
        self.penalizations = penalizations

    def get_fitness(self, simulResult, candidate, **kwargs):
        sum = 0
        for v in candidate.values():
            if v == 0:
                sum += self.penalizations['KO']
            elif v < 1:
                sum += self.penalizations['UE']
            else:
                sum += self.penalizations['OE']
        return sum / len(candidate)

    def required_simulations(self):
        """
        If the evaluation function requires a pre-simulation to compute fitness values
        """
        return []

    def short_str(self):
        return "ModificationType"

    def method_str(self):
        return "ModificationType"


class MolecularWeight(PhenotypeEvaluationFunction):
    """Minimizes the sum of molecular weights of the products of a set of reactions (g/gDW/h).
    """

    def __init__(self, reactions, maximize=False, **kwargs):
        super(MolecularWeight, self).__init__(maximize=maximize, worst_fitness=np.inf)
        self.reactions = reactions
        self.method = kwargs.get('method', SimulationMethod.pFBA)
        # sum of molar masses of product compounds for the reactions
        self.__mw = None

    def compute_rxnmw(self, model):
        from ..util.constants import atomic_weights
        self.__mw = {}
        simulator = get_simulator(model)
        for rx in self.reactions:
            p = simulator.get_products(rx)
            if not p:
                p = simulator.get_substrates(rx)
            rmw = 0
            for m, v in p.items():
                elem = simulator.metabolite_elements(m)
                w = 0
                for e, n in elem.items():
                    try:
                        w += atomic_weights[e] * n
                    except:
                        pass

                rmw += abs(v) * w

            self.__mw[rx] = rmw

    def get_fitness(self, simul_results, candidate, **kwargs):
        try:
            sim = simul_results[self.method]
        except Exception:
            sim = None

        if sim.status not in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
            return self.no_solution

        if self.__mw is None:
            self.compute_rxnmw(sim.model)

        fitness = 0
        for rx in self.reactions:
            fitness += self.__mw[rx] * abs(sim.fluxes[rx])

        if fitness > 0:
            return fitness * 0.001
        else:
            return self.no_solution

    def required_simulations(self):
        """
        If the evaluation function requires a pre-simulation to compute fitness values
        """
        return [self.method]

    def short_str(self):
        return "Molecular Weight"

    def method_str(self):
        return "MW"



from typing import Union, Dict, Any


def solution_decode(solution: Union[int, bool], decoder: Dict[Any, int] = None) -> int:
    """
    Decode a solution from expression or symbolic evaluation.
    Solution decoding is sometimes not straightforward due to python bitwise evaluations

    :param solution: a boolean or int result
    :param decoder: a custom dictionary for decoding the solution (key) into a given output (value)
    :return: it returns a solution decoded into the integer 0 or 1
    """

    if not decoder:
        decoder = {True: 1,
                   False: 0,
                   1: 1,
                   0: 0,
                   -1: 1,
                   -2: 0
                   }

    return decoder.get(solution, solution)


def _walk(symbolic, reverse=False):
    """
    Internal use!
    Helping function to walk through an algebra expression implemented by the symbolic objects.
    This function is heavily inspired by the sympy method preorder_traversal and postorder_traversal.
    Credits to sympy contributors: https://github.com/sympy/sympy

    :param symbolic: symbolic-like object
    :param reverse: order of the traversal
    :return:
    """
    if reverse:

        for child in symbolic.variables:
            for subtree in _walk(child, reverse):
                yield subtree

        yield symbolic

    else:

        yield symbolic

        for child in symbolic.variables:
            yield from _walk(child, reverse)



from collections import namedtuple
from dataclasses import dataclass
from typing import TYPE_CHECKING, Tuple, Union, Dict, Callable

from math import log, exp

from mewpy.germ.algebra import And, Or
from mewpy.solvers.solution import Status
from mewpy.util.constants import ModelConstants

if TYPE_CHECKING:
    from mewpy.solvers import Solution
    from mewpy.germ.lp import LinearProblem
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def decode_solver_solution(solution: 'Solution') -> Tuple[float, str]:
    """
    It decodes the solution of the solver and returns the objective value.

    :param solution: the solution of the solver
    :return: the objective value and the status of the solution
    """
    sol_status = solution.status.value

    if solution.status == Status.OPTIMAL:
        sol_f_obj = solution.fobj

    elif solution.status == Status.UNBOUNDED or solution.status == Status.INF_OR_UNB:

        sol_f_obj = ModelConstants.REACTION_UPPER_BOUND

    else:
        sol_f_obj = 0.0

    return sol_f_obj, sol_status


def run_method_and_decode(method: 'LinearProblem',
                          objective: Union[str, Dict[str, float]] = None,
                          constraints: Dict[str, Tuple[float, float]] = None,
                          **kwargs) -> Tuple[float, str]:
    """
    It runs a method and decodes the objective value and status returned by the solver.
    :param method: the method to be run
    :param objective: an alternative temporary objective function
    :param constraints: alternative temporary constraints
    :param kwargs: additional arguments to be passed to the method
    :return: the objective value and the status of the solution
    """
    solver_kwargs = {'get_values': False}

    if objective:
        if hasattr(objective, 'keys'):
            solver_kwargs['linear'] = objective.copy()
        else:
            solver_kwargs['linear'] = {str(objective): 1.0}

    if constraints:
        solver_kwargs['constraints'] = constraints

    if 'minimize' in kwargs:
        solver_kwargs['minimize'] = kwargs['minimize']

    solution = method.optimize(to_solver=True, solver_kwargs=solver_kwargs, **kwargs)
    objective_value, status = decode_solver_solution(solution=solution)
    return objective_value, status


# ---------------------------------
# CoRegFlux utils
# ---------------------------------
CoRegMetabolite = namedtuple('CoRegMetabolite', ('id', 'concentration', 'exchange'))
CoRegBiomass = namedtuple('CoRegBiomass', ('id', 'biomass_yield'))


@dataclass
class CoRegResult:
    objective_value: float = None
    values: Dict[str, float] = None
    metabolites: Dict[str, CoRegMetabolite] = None
    biomass: CoRegBiomass = None
    state: Dict[str, float] = None
    constraints: Dict[str, Tuple[float, float]] = None


def build_metabolites(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                      metabolites: Dict[str, float]) -> Dict[str, CoRegMetabolite]:
    res = {}
    for metabolite, concentration in metabolites.items():
        exchange = model.get(metabolite).exchange_reaction.id

        res[metabolite] = CoRegMetabolite(id=metabolite,
                                          concentration=concentration,
                                          exchange=exchange)
    return res


def build_biomass(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                  biomass: float) -> CoRegBiomass:
    variable = next(iter(model.objective))
    return CoRegBiomass(id=variable.id, biomass_yield=biomass)


def concentration_to_lb(concentration,
                        biomass,
                        time_step):
    return concentration / (biomass * time_step)


def lb_soft_plus(soft_plus, coefficient):
    return -1 * log(1 + exp(soft_plus + abs(coefficient)))


def ub_soft_plus(soft_plus, coefficient):
    return log(1 + exp(soft_plus + coefficient))


def euler_step_biomass(old_biomass_yield, growth_rate, time_step):
    return old_biomass_yield * exp(growth_rate * time_step)


def euler_step_metabolites(metabolite_concentration,
                           metabolite_rate,
                           old_biomass_yield,
                           growth_rate,
                           time_step):
    next_concentration = metabolite_rate / growth_rate * old_biomass_yield * (1 - exp(growth_rate * time_step))
    return metabolite_concentration - next_concentration


def biomass_yield_to_rate(biomass):
    return 1 * biomass


def metabolites_constraints(constraints: Dict[str, Tuple[float, float]],
                            metabolites: Dict[str, CoRegMetabolite],
                            biomass: CoRegBiomass,
                            time_step: float):
    constraints = constraints.copy()
    for metabolite in metabolites.values():

        next_lb = concentration_to_lb(concentration=metabolite.concentration,
                                      biomass=biomass.biomass_yield,
                                      time_step=time_step)

        rxn = metabolite.exchange

        if next_lb < ModelConstants.TOLERANCE:
            next_lb = 0

        prev_lb, prev_ub = constraints[rxn]

        if prev_lb == 0:
            next_lb = -next_lb

        else:

            if abs(prev_lb) < abs(next_lb):
                next_lb = prev_lb

            else:
                next_lb = -abs(next_lb)

        constraints[rxn] = (next_lb, prev_ub)

    return constraints


def continuous_gpr(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                   state: Dict[str, float],
                   scale: bool = False):
    operators = {And: min, Or: max}

    states = {}
    for reaction in model.yield_reactions():

        if reaction.gpr.is_none:
            continue

        if not set(reaction.genes).issubset(state):
            continue

        states[reaction.id] = reaction.gpr.evaluate(values=state, operators=operators, missing_value=0)

    if scale:
        _max_state = max(states.values())

        return {key: round((val / _max_state), 6) * 1000 for key, val in states.items()}

    return states


def gene_state_constraints(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                           constraints: Dict[str, Tuple[float, float]],
                           state: Dict[str, float],
                           soft_plus: float = 0,
                           tolerance: float = ModelConstants.TOLERANCE,
                           scale: bool = False):
    constraints = constraints.copy()

    reactions_state = continuous_gpr(model=model, state=state, scale=scale)
    # find the reactions which lower bound was changed by the rules
    # find the reactions which upper bound was changed by the rules
    # evaluate the soft plus over the bounds computed using the continuous version of the gpr rules
    # gene_state_bounds does not contain reactions without gpr rules
    for reaction, coefficient in reactions_state.items():

        old_lb, old_ub = constraints[reaction]

        if coefficient > tolerance:

            if abs(old_lb) > tolerance:
                new_lb = lb_soft_plus(soft_plus=soft_plus, coefficient=coefficient)
            else:
                new_lb = old_lb

            if old_ub > tolerance:
                new_ub = ub_soft_plus(soft_plus=soft_plus, coefficient=coefficient)
            else:
                new_ub = old_ub

        else:
            new_lb = old_lb
            new_ub = old_ub

        constraints[reaction] = (new_lb, new_ub)

    return constraints


def system_state_update(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                        flux_state: Dict[str, float],
                        metabolites: Dict[str, CoRegMetabolite],
                        biomass: CoRegBiomass,
                        time_step: float,
                        biomass_fn: Callable = None) -> Tuple[CoRegBiomass, Dict[str, CoRegMetabolite]]:

    growth_rate = flux_state[biomass.id]
    old_biomass_yield = biomass.biomass_yield

    if biomass_fn is not None:
        growth_rate = biomass_fn(growth_rate)

    if growth_rate == 0:
        growth_rate = ModelConstants.TOLERANCE

    biomass_yield = euler_step_biomass(old_biomass_yield=old_biomass_yield,
                                       growth_rate=growth_rate,
                                       time_step=time_step)

    next_biomass = build_biomass(model, biomass_yield)

    next_metabolites = {}

    for met_id, met in metabolites.items():

        concentration = euler_step_metabolites(metabolite_concentration=met.concentration,
                                               metabolite_rate=flux_state[met.exchange],
                                               old_biomass_yield=old_biomass_yield,
                                               growth_rate=growth_rate,
                                               time_step=time_step)

        if concentration < 0:
            concentration = 0

        next_metabolites[met_id] = CoRegMetabolite(id=met_id,
                                                   concentration=concentration,
                                                   exchange=met.exchange)

    return next_biomass, next_metabolites



from typing import Any, Union, Type, TYPE_CHECKING, List, Set, Tuple, Dict, Iterable

from mewpy.util.history import HistoryManager, recorder
from mewpy.germ.models.serialization import serialize, Serializer

# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
    from mewpy.germ.algebra import Expression, Symbolic

    from .gene import Gene
    from .interaction import Interaction
    from .metabolite import Metabolite
    from .reaction import Reaction
    from .regulator import Regulator
    from .target import Target


class MetaVariable(type):
    """
    The MetaVariable class is used to dynamically create Variables classes.
    Variables instances are created from static classes representing the different information for a given variable.

    However, integrated Metabolic-Regulatory models may have different types of variables, such as genes, reactions,
    metabolites, regulators, targets, interactions, but also metabolite-regulator, gene-target, etc.
    Therefore, to avoid the duplication of certain variables, a Variable factory is used to dynamically create
    Variable classes based on the multiple types of a single variable.

    The Variable factory is a class that is the interface to create these multi-type variables.
    It also manages all variables by implementing base attributes and methods.

    The MetaVariable class creates a dynamic Variable class as follows:
        1. It collects all static base classes to be included into the dynamic Variable class
        2. It collects the variable type from the base classes or from the variable_type attribute
        3. It sets the name of the dynamic Variable class using the name of the base classes
        4. It updates the variable_type attribute of the dynamic Variable class with all types of the base classes
        5. It creates the dynamic Variable class having all attributes and methods of the base classes
        6. It adds polymorphic constructors to the dynamic Variable class based on the types of the base classes
        7. It adds type checkers to the dynamic Variable class based on the types of the base classes
    """
    factories = {}

    def __new__(mcs, name, bases, attrs, **kwargs):

        # if it is the variable factory, only registration is done
        factory = kwargs.get('factory', False)

        if factory:
            cls = super(MetaVariable, mcs).__new__(mcs, name, bases, attrs)

            MetaVariable.factories[name] = cls

            return cls

        # Dynamic typing being used. In this case, a proper name and variable type must be provided
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            names = [base.variable_type for base in bases]

            name = ''.join([name.title() for name in names])
            name += 'Variable'

            kwargs['variable_type'] = '-'.join(names)

        # The variable type is always added to the subclasses. If it is not given upon subclass creation,
        # the subclass name is to be used
        variable_type = kwargs.get('variable_type', name.lower())
        attrs['variable_type'] = variable_type

        return super(MetaVariable, mcs).__new__(mcs, name, bases, attrs)

    def __init__(cls, name, bases, attrs, **kwargs):

        super().__init__(name, bases, attrs)

        factory = kwargs.get('factory', False)
        if factory:
            # collection of all attributes that must be serialized for a particular class or subclass
            attributes = cls.get_serializable_attributes(attrs)

            attrs['_attributes_registry']['variable'] = attributes

            # Skip further building of the Variable factory
            return

        # Dynamic typing being used. In this case, all children have already been constructed, so everything can be
        # skipped
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            return

        # Several attributes and methods must be added automatically to the Variable factory children based on the
        # variable type.
        variable_type = attrs['variable_type']

        for base in bases:

            factory = MetaVariable.factories.get(base.__name__)

            if factory and hasattr(cls, 'variable_type'):

                # if some class inherits from the Variable factory, it must be registered in the factory for type
                # checking
                if cls.variable_type not in factory.get_registry():
                    raise TypeError(f'{cls.variable_type} does not inherit from Variable')

                # If set otherwise upon class creation, all subclasses of the Variable factory will contribute to
                # their parent factory with an alternative initializer. The polymorphic constructor, e.g. from_{
                # variable_type}, is added to the Variable factory
                constructor = kwargs.get('constructor', True)

                if constructor:
                    # noinspection PyProtectedMember
                    var_type_initializer = Variable._from(cls)
                    setattr(factory, f'from_{variable_type}', var_type_initializer)

                # If set otherwise upon class creation, all subclasses of the Variable factory will contribute to
                # their parent factory with a declared type checker. The type checker, e.g. is_{variable_type},
                # is added to the Variable factory
                checker = kwargs.get('checker', True)

                if checker:
                    # noinspection PyProtectedMember
                    var_type_checker = Variable._is(variable_type)
                    setattr(factory, f'is_{variable_type}', var_type_checker)

                # collection of all attributes that must be serialized for a particular class or sub class
                attributes = cls.get_serializable_attributes(attrs)

                Variable.register_attributes(attributes, cls)

    @staticmethod
    def get_serializable_attributes(attrs: Dict[str, Any]) -> Dict[str, Tuple[str, str, str]]:
        """
        Collects all attributes that must be serialized for a particular class or subclass.

        All variables have property descriptors marked with the @serialize decorator. This decorator marks attributes
        for serialization, deserialization and pickle serialization/deserialization. The decorator also provides
        information about the attribute name.

        If serialization, deserialization or pickle serialization/deserialization is not desired, the attribute must
        be marked with the @serialize decorator with the corresponding argument set to None.

        :param attrs: the attributes of the class or subclass
        :return: a dictionary with the attribute name as key and a tuple with the attribute name for serialization,
        deserialization and pickle
        """

        attributes = {}

        for name, method in attrs.items():

            if hasattr(method, 'fget'):

                if hasattr(method.fget, 'serialize') and hasattr(method.fget, 'deserialize') and hasattr(method.fget,
                                                                                                         'pickle'):
                    attributes[name] = (method.fget.serialize, method.fget.deserialize, method.fget.pickle)

        return attributes


class Variable(Serializer, metaclass=MetaVariable, factory=True):
    """
    The Variable class is the base class for all variables and the factory for all variables.
    It is the interface to create variables of different types providing several base attributes and methods.

    Base attributes:
        - types: the types of the variable
        - id: the identifier of the variable
        - name: the name of the variable
        - aliases: the aliases of the variable
        - model: the model to which the variable belongs
        - contexts: the contexts in which the variable is used
        - history: the history of the variable

    Base methods:
        - factory: the factory for dynamic variable types
        - from_types: creates a variable from a list of types
        - copy: creates a copy of the variable
        - deepcopy: creates a deep copy of the variable
        - undo: undoes the last change of the variable
        - redo: redoes the last change of the variable
        - reset: resets the variable to its initial state
        - restore: restores the variable to a previous state
        - update: updates the variable with a new information
        - from polymorphic constructors: creates a variable from a specific type
        - is polymorphic type checkers: checks if a variable is of a specific type

    Serialization:
        - to_dict: serializes the variable to a dictionary
        - from_dict: deserializes the variable from a dictionary

    The Variable class holds the registry of all potential variables.
    """

    # -----------------------------------------------------------------------------
    # Factory management
    # -----------------------------------------------------------------------------
    _registry = {}
    _attributes_registry = {'variable': {}}

    def __init_subclass__(cls, **kwargs):
        """
        This method is called when a subclass is created. It is used to register the subclass in the registry of the
        Variable factory.

        Internal use only.
        :param kwargs:
        :return:
        """

        super(Variable, cls).__init_subclass__(**kwargs)

        # the child type
        variable_type = getattr(cls, 'variable_type', cls.__name__.lower())

        cls.register_type(variable_type, cls)

    @staticmethod
    def get_registry() -> Dict[str, Type['Variable']]:
        """
        Returns the registry of the Variable factory.

        Internal use only.
        :return: the registry of the Variable factory as a dictionary
        """
        return Variable._registry.copy()

    @staticmethod
    def get_attributes_registry() -> Dict[str, Dict[str, Tuple[str, str, str]]]:
        """
        Returns the registry of the attributes of the Variable factory.

        Internal use only.
        :return: the registry of the attributes of the Variable factory as a dictionary
        """
        return Variable._attributes_registry.copy()

    @staticmethod
    def register_type(variable_type: str, child: Type['Variable']):
        """
        Registers a child in the registry of the Variable factory.

        Internal use only.
        :param variable_type: the type of the child
        :param child: the child to be registered
        :return:
        """
        Variable._registry[variable_type] = child

    # -----------------------------------------------------------------------------
    # Serialization
    # -----------------------------------------------------------------------------
    @staticmethod
    def register_attributes(attributes, child):
        """
        Registers the attributes of a child class in the attributes registry of the Variable factory. This is useful for
        serialization and deserialization.

        Internal use only.
        :param attributes: the attributes of the child class
        :param child: the child class
        :return:
        """

        if hasattr(child, 'variable_type'):
            Variable._attributes_registry[child.variable_type] = attributes

        elif child is Variable:
            Variable._attributes_registry['variable'] = attributes

    @property
    def attributes(self) -> Dict[str, Tuple[str, str, str]]:
        """
        Returns the attributes of the variable. This is useful for serialization and deserialization.

        Internal use only.
        :return: the attributes of the variable as a dictionary
        """
        class_attributes = self.get_attributes_registry()

        attributes = class_attributes['variable'].copy()

        for variable_type in self.types:

            for name, (serialize_name, deserialize_name, pickle_name) in class_attributes[variable_type].items():
                attributes[name] = (serialize_name, deserialize_name, pickle_name)

        return attributes

    # -----------------------------------------------------------------------------
    # Factory polymorphic constructor
    # -----------------------------------------------------------------------------
    @classmethod
    def factory(cls, *args: str) -> Union[Type['Variable'],
                                          Type['Gene'],
                                          Type['Interaction'],
                                          Type['Metabolite'],
                                          Type['Reaction'],
                                          Type['Regulator'],
                                          Type['Target']]:
        """
        It creates a dynamic Variable class from a list of types.
        All types must be registered in the Variable factory.

        Example:
        >>> Variable.factory('gene', 'target')
        <class 'mewpy.germ.variables.GeneTargetVariable'>

        :param args: the types of the variable
        :return: the variable type
        """

        if not args:
            args = ()

        registry = cls.get_registry()

        types = tuple([registry[name] for name in args])

        if len(types) == 1:
            return types[0]

        _Variable = MetaVariable('Variable', types, {}, dynamic=True)

        return _Variable

    # -----------------------------------------------------------------------------
    # Factory polymorphic initializer
    # -----------------------------------------------------------------------------
    @classmethod
    def from_types(cls, types: Iterable[str], **kwargs) -> Union['Variable',
                                                                 'Gene',
                                                                 'Interaction',
                                                                 'Metabolite',
                                                                 'Reaction',
                                                                 'Regulator',
                                                                 'Target']:
        """
        It creates a variable instance from a list of types and a dictionary of attributes.
        All types must be registered in the Variable factory so the factory can create a dynamic Variable class.
        Then, the variable is initialized with the provided attributes.

        Example:
        >>> Variable.from_types(['gene', 'target'], id='GENE_1', name='Gene 1')
        <mewpy.germ.variables.GeneTargetVariable object at 0x7f8b8c0b7a90>

        :param types: the types of the variable
        :param kwargs: the attributes to initialize the variable
        :return: the variable instance
        """
        VariableType = cls.factory(*types)
        return VariableType(**kwargs)

    # -----------------------------------------------------------------------------
    # Factory helper methods
    # -----------------------------------------------------------------------------
    @staticmethod
    def _from(cls):
        """
        Method to be added to the subclasses of Variable to create a variable from a specific type.

        Internal use only.
        :param cls: the class of the variable
        :return: the method to be added to the subclasses of Variable
        """

        def from_(identifier, **kwargs):
            return cls(identifier, **kwargs)

        return from_

    @staticmethod
    def _is(variable_type):
        """
        Method to be added to the subclasses of Variable to check if a variable is of a specific type.

        Internal use only.
        :param variable_type: the type of the variable
        :return: the method to be added to the subclasses of Variable
        """

        def is_(self):
            return self.is_a(variable_type)

        return is_

    # -----------------------------------------------------------------------------
    # Base initializer
    # -----------------------------------------------------------------------------

    def __init__(self,
                 identifier: Any,
                 name: str = None,
                 aliases: Set[str] = None,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):

        """
        Variable is the most base type of all variables in MEWpy,
        such as gene, interaction, metabolite, reaction, regulator and target, among others.
        See also variables.Gene, variables.Interaction, variables.Metabolite, variables.Reaction,
        variables.Regulator and variables.Target for concrete implementations of a variable type.

        The variable is also the factory for several variable types.
        The Variable factory provides all batteries to create new variable types
        and to manage the variables types' inheritance.

        The factory type assists with systematic tasks such as attributes and containers serialization,
        history management, among others.

        For that, attributes, containers and inheritance are registered in the factory. The MetaVariable manages
        both the factory and the variable types derived from the factory.

        :param identifier: identifier, e.g. b0001
        :param name: the name of the variable
        :param aliases: aliases for this variable
        :param model: the model to which this variable belongs. Since the default is None, variables do not need to
        be directly associated with a model
        """

        self._check_inheritance()

        if not aliases:
            aliases = set()

        if not model:
            model = None

        if not name:
            name = ''

        self._id = identifier
        self._aliases = aliases
        self._model = model
        self._name = name

        # history
        self._contexts = []
        self._history = HistoryManager()

        self._model_ref = 0

    def _check_inheritance(self):
        """
        It checks if the class is a subclass of Variable and if it is not, it raises an error.
        :return:
        """

        registry = self.get_registry()

        for variable_type in self.types:
            if variable_type not in registry:
                raise ValueError(f'{variable_type} is not registered as subclass of {self.__class__.__name__}')

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    def _variable_to_html(self):
        """
        It returns an HTML dict representation of the variable.
        """
        html_dict = {'Identifier': self.id,
                     'Name': self.name,
                     'Aliases': ', '.join(self.aliases),
                     'Model': self.model.id if self.model else None,
                     'Types': ', '.join(self.types)}
        return html_dict

    def _repr_html_(self):
        """
        It returns a html representation.
        """
        html_dict = self._variable_to_html()

        for type_ in self.types:
            method = getattr(self, f'_{type_}_to_html', lambda: {})
            html_dict.update(method())

        html_representation = ''
        for key, value in html_dict.items():
            html_representation += f'<tr><th>{key}</th><td>{value}</td></tr>'

        return f"""
            <table>
                {html_representation}
            </table>
        """

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------
    @serialize('types', None)
    @property
    def types(self) -> Set[str]:
        """
        It returns the types of the variable.
        :return: the types of the variable as a set of strings
        """
        return set()

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('id', None, '_id')
    @property
    def id(self) -> Any:
        """
        It returns the identifier of the variable.
        :return: the identifier of the variable
        """
        return self._id

    @serialize('name', 'name', '_name')
    @property
    def name(self) -> str:
        """
        It returns the name of the variable.
        :return: the name of the variable as a string
        """
        return self._name

    @serialize('aliases', 'aliases', '_aliases')
    @property
    def aliases(self) -> Set[str]:
        """
        It returns the aliases of the variable.
        :return: the aliases of the variable as a set of strings
        """
        return self._aliases

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        """
        It returns the model to which the variable belongs.
        Note that variables do not need to be directly associated with a model.
        :return: the model to which the variable belongs
        """
        return self._model

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @name.setter
    @recorder
    def name(self, value: str):
        """
        It sets the name of the variable.
        :param value: the name of the variable
        :return:
        """
        if not value:
            value = ''

        self._name = value

    @aliases.setter
    @recorder
    def aliases(self, value: Set[str]):
        """
        It sets the aliases of the variable.
        :param value: the aliases of the variable
        :return:
        """
        if not value:
            value = set()

        self._aliases = value

    @model.setter
    def model(self, value: Union['Model', 'MetabolicModel', 'RegulatoryModel']):
        """
        It sets the model to which the variable belongs.
        This setter does not perform any check or action in the model.

        :param value: the model to which the variable belongs
        :return:
        """
        if not value:
            value = None

        self._model = value

    # -----------------------------------------------------------------------------
    # History manager
    # -----------------------------------------------------------------------------

    # The history manager follows the command pattern
    @property
    def contexts(self) -> List[HistoryManager]:
        """
        It returns the contexts of the variable.
        :return: the contexts of the variable as a list of HistoryManager
        """

        if self.model:
            return self.model.contexts

        return self._contexts

    @property
    def history(self) -> HistoryManager:
        """
        It returns the history manager of the variable.
        If the variable is within a context, the history manager of the last context is returned.
        :return: the history manager of the variable
        """
        if self.contexts:
            return self.contexts[-1]

        if self.model:
            return self.model.history

        return self._history

    def undo(self):
        """
        It undoes the last action performed on the variable.
        :return:
        """
        self.history.undo()

    def redo(self):
        """
        It redoes the last action performed on the variable.
        :return:
        """
        self.history.redo()

    def clean_history(self):
        """
        It cleans the history of the variable.
        :return:
        """
        self._history = HistoryManager()

    def clean_context(self):
        """
        It cleans the context of the variable.
        :return:
        """
        self.contexts[:-1] = HistoryManager()

    def reset(self):
        """
        It resets the variable to its initial state.
        :return:
        """
        self.history.reset()

    def restore(self):
        """
        It restores the variable to its last state.
        :return:
        """
        self.history.restore()

    def __enter__(self):
        """
        It enters the context of the variable.
        It creates a new context and adds it to the list of contexts.
        All changes performed within this context will be reset when the context is exited.

        :return: the variable itself
        """
        self.contexts.append(HistoryManager())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        It exits the context of the variable performing a history reset.

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        context = self.contexts.pop()
        context.reset()

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def update(self,
               name: str = None,
               aliases: Set[str] = None,
               model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):
        """
        It updates the attributes of the variable.
        :param name: the name of the variable
        :param aliases: the aliases of the variable
        :param model: the model to which the variable belongs
        :return:
        """
        if name is not None:
            self.name = name

        if aliases is not None:
            self.aliases.update(set(aliases))

        if model is not None:
            self.model = model

    # -----------------------------------------------------------------------------
    # Abstract implementations
    # -----------------------------------------------------------------------------

    # The following polymorphic initializers are just registered here to avoid type checking errors
    @classmethod
    def from_gene(cls,
                  identifier: Any,
                  name: str = None,
                  aliases: set = None,
                  model: 'Model' = None,
                  coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                  active_coefficient: Union[int, float] = None,
                  reactions: Dict[str, 'Reaction'] = None) -> 'Gene':
        """
        It creates a gene from the given parameters.
        :param identifier: the identifier of the gene
        :param name: the name of the gene
        :param aliases: the aliases of the gene
        :param model: the model to which the gene belongs
        :param coefficients: the coefficients of the gene
        :param active_coefficient: the default coefficient of the gene
        :param reactions: the reactions associated with the gene
        :return: a gene object
        """
        ...

    @classmethod
    def from_interaction(cls,
                         identifier: Any,
                         name: str = None,
                         aliases: set = None,
                         model: 'Model' = None,
                         target: 'Target' = None,
                         regulatory_events: Dict[Union[float, int], 'Expression'] = None) -> 'Interaction':
        """
        It creates an interaction from the given parameters.
        :param identifier: the identifier of the interaction
        :param name: the name of the interaction
        :param aliases: the aliases of the interaction
        :param model: the model to which the interaction belongs
        :param target: the target of the interaction
        :param regulatory_events: the regulatory events associated with the interaction
        :return: an interaction object
        """
        ...

    @classmethod
    def from_metabolite(cls,
                        identifier: Any,
                        name: str = None,
                        aliases: set = None,
                        model: 'Model' = None,
                        charge: int = None,
                        compartment: str = None,
                        formula: str = None,
                        reactions: Dict[str, 'Reaction'] = None) -> 'Metabolite':
        """
        It creates a metabolite from the given parameters.
        :param identifier: the identifier of the metabolite
        :param name: the name of the metabolite
        :param aliases: the aliases of the metabolite
        :param model: the model to which the metabolite belongs
        :param charge: the charge of the metabolite
        :param compartment: the compartment of the metabolite
        :param formula: the formula of the metabolite
        :param reactions: the reactions associated with the metabolite
        :return: a metabolite object
        """
        ...

    @classmethod
    def from_reaction(cls,
                      identifier: Any,
                      name: str = None,
                      aliases: set = None,
                      model: 'Model' = None,
                      bounds: Tuple[Union[float, int], Union[float, int]] = None,
                      stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                      gpr: 'Expression' = None) -> 'Reaction':
        """
        It creates a reaction from the given parameters.
        :param identifier: the identifier of the reaction
        :param name: the name of the reaction
        :param aliases: the aliases of the reaction
        :param model: the model to which the reaction belongs
        :param bounds: the bounds of the reaction
        :param stoichiometry: the stoichiometry of the reaction
        :param gpr: the gene-protein-reaction association of the reaction
        :return: a reaction object
        """
        ...

    @classmethod
    def from_regulator(cls,
                       identifier: Any,
                       name: str = None,
                       aliases: set = None,
                       model: 'Model' = None,
                       coefficients: Set[Union[float, int]] = None,
                       active_coefficient: Union[float, int] = None,
                       interactions: Dict[str, 'Interaction'] = None) -> 'Regulator':
        """
        It creates a regulator from the given parameters.
        :param identifier: the identifier of the regulator
        :param name: the name of the regulator
        :param aliases: the aliases of the regulator
        :param model: the model to which the regulator belongs
        :param coefficients: the coefficients of the regulator
        :param active_coefficient: the default coefficient of the regulator
        :param interactions: the interactions associated with the regulator
        :return: a regulator object
        """
        ...

    @classmethod
    def from_target(cls,
                    identifier: Any,
                    name: str = None,
                    aliases: set = None,
                    model: 'Model' = None,
                    coefficients: Set[Union[float, int]] = None,
                    active_coefficient: Union[float, int] = None,
                    interaction: 'Interaction' = None) -> 'Target':
        """
        It creates a target from the given parameters.
        :param identifier: the identifier of the target
        :param name: the name of the target
        :param aliases: the aliases of the target
        :param model: the model to which the target belongs
        :param coefficients: the coefficients of the target
        :param active_coefficient: the default coefficient of the target
        :param interaction: the interaction associated with the target
        :return: a target object
        """
        ...

    # The following type checkers are just registered here to avoid type checking errors
    def is_gene(self) -> bool:
        """
        It checks whether the variable is a gene.
        :return: True if the variable is a gene, False otherwise
        """
        ...

    def is_interaction(self) -> bool:
        """
        It checks whether the variable is an interaction.
        :return: True if the variable is an interaction, False otherwise
        """
        ...

    def is_metabolite(self) -> bool:
        """
        It checks whether the variable is a metabolite.
        :return: True if the variable is a metabolite, False otherwise
        """
        ...

    def is_reaction(self) -> bool:
        """
        It checks whether the variable is a reaction.
        :return: True if the variable is a reaction, False otherwise
        """
        ...

    def is_regulator(self) -> bool:
        """
        It checks whether the variable is a regulator.
        :return: True if the variable is a regulator, False otherwise
        """
        ...

    def is_target(self) -> bool:
        """
        It checks whether the variable is a target.
        :return: True if the variable is a target, False otherwise
        """
        ...

    # -----------------------------------------------------------------------------
    # Type checker
    # -----------------------------------------------------------------------------

    def is_a(self, variable_type) -> bool:
        """
        It checks whether the variable is of the given type.
        :param variable_type: the type to check
        :return: True if the variable is of the given type, False otherwise
        """
        if variable_type in self.types:
            return True

        return False


# -----------------------------------------------------------------------------
# Helper methods
# -----------------------------------------------------------------------------
def variables_from_symbolic(symbolic: 'Symbolic',
                            types: Union[Set[str], List[str], Tuple[str]],
                            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None) -> Dict[str, 'Variable']:
    """
    It returns the variables of the given type from the given symbolic expression.
    :param symbolic: the symbolic expression to parse and extract the variables from
    :param types: the types of the variables to return
    :param model: the model to which the variables belong
    :return: the variables of the given type from the given symbolic expression
    """
    variables = {}

    for symbol in symbolic.atoms(symbols_only=True):

        if symbol.name in variables:
            continue

        if model:

            variable = model.get(symbol.name)

            if variable:

                for variable_type in types:

                    if not variable.is_a(variable_type):
                        raise TypeError(f'{symbol.name} is not a {variable_type} in model {model.id}')

        else:
            variable = Variable.from_types(identifier=symbol.name,
                                           types=types,
                                           model=model)

        variables[variable.id] = variable

    return variables


def build_variable(types: Iterable[str], kwargs: Dict[str, Any]) -> Union['Variable',
                                                                          'Gene',
                                                                          'Interaction',
                                                                          'Metabolite',
                                                                          'Reaction',
                                                                          'Regulator',
                                                                          'Target']:
    """
    It builds a variable from the given types and keyword arguments. Check the `Variable.from_types()` method for more
    details.
    :param types: the types of the variable
    :param kwargs: the keyword arguments to build the variable
    :return: a new variable of the given types
    """
    return Variable.from_types(types, **kwargs)



"""odespy ODE Solver
"""
from .ode import ODEMethod, SolverConfigurations, ODESolver
import numpy as np
import odespy

methods = {
    ODEMethod.LSODA: odespy.Lsoda,
    ODEMethod.LSODAR: odespy.Lsodar,
    ODEMethod.LSODE: odespy.Lsode,
    ODEMethod.HEUN: odespy.Heun,
    ODEMethod.EULER: odespy.Euler,
    ODEMethod.RK4: odespy.RK4,
    ODEMethod.DORMAN_PRINCE: odespy.DormandPrince,
    ODEMethod.RKFehlberg: odespy.RKFehlberg,
    ODEMethod.Dopri5: odespy.Dopri5,
    # ODEMethod.Dop853: odespy.Dop853,
    ODEMethod.Vode: odespy.Vode,
    ODEMethod.Radau: odespy.Radau5,
    ODEMethod.AdamsBashMoulton2: odespy.AdamsBashMoulton2,
    ODEMethod.AdamsBashforth2: odespy.AdamsBashforth2
}


class ODESpySolver(ODESolver):
    """
    ODE solver method implemented on odespy package.
    """

    def __init__(self, func, method):
        self.func = func
        if method in methods.keys():
            self.method = method
        else:
            raise ValueError(f'Method {method} is unavailable.')

        self.initial_condition = None

    def set_initial_condition(self, initial_condition):
        self.initial_condition = initial_condition

    def solve(self, y0, t_points, **kwargs):
        """
        Solves the ODE
        """
        def f(u, t):
            return self.func(t, u)

        try:
            if self.method == ODEMethod.AdamsBashforth2:
                solver = methods[self.method](f, method='bdf')
            else:
                solver = methods[self.method](f)
            
            # update default parameters
            time_points = t_points
            solver.atol = SolverConfigurations.ABSOLUTE_TOL
            solver.rtol = SolverConfigurations.RELATIVE_TOL
            solver.set_initial_condition(y0)
            y, t = solver.solve(time_points)
            C = [c[-1] for c in y]
            return C, t, y
        except Exception as e:
            print(e)
            raise(Exception)




from .operators import (uniform_crossover_OU, grow_mutation_OU, shrink_mutation,
                        single_mutation_OU, uniform_crossover_KO, grow_mutation_KO,
                        single_mutation_KO)

from ..settings import get_default_population_size


def get_population_size():
    size = get_default_population_size()
    return size


OU = {
    'variators': [uniform_crossover_OU,
                  grow_mutation_OU,
                  shrink_mutation,
                  single_mutation_OU]
}

KO = {
    'variators': [uniform_crossover_KO,
                  grow_mutation_KO,
                  shrink_mutation,
                  single_mutation_KO]
}

PARAMETERS = {'gs_mutation_rate': 0.1,
              'mutation_rate': 0.1,
              'crossover_rate': 0.9,
              'tournament_size': 7,
              }



from typing import Union, TYPE_CHECKING, Dict, Callable, Any, Type, Sequence
from warnings import warn

import pandas as pd

from mewpy.germ.algebra import Symbolic

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def regulatory_truth_table(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                           interactions: Sequence[str] = None,
                           initial_state: Dict[str, float] = None,
                           strategy: str = 'max',
                           operators: Union[Dict[Type[Symbolic], Callable], Dict[Type[Symbolic], Any]] = None,
                           decoder: dict = None) -> pd.DataFrame:
    """
    The regulatory truth table of a regulatory model contains the evaluation of all regulatory events.
    RegulatoryModel's interactions are evaluated using an initial state or regulators coefficients.
    The results are stored in a dictionary or a pandas DataFrame.

    :param model: A regulatory or metabolic-regulatory model to be simulated
    :param interactions: A list of interactions to be evaluated. If None, all interactions are evaluated (default).
    :param initial_state: A dictionary with the initial state of the model. If None, the default initial state is used
    :param strategy: The truth table can be calculated using the maximum or minimum value
    in the variables' coefficients. Otherwise, the truth table is calculated using all variables' coefficients.
    :param operators: A dictionary with custom operators to be used in the evaluation of the regulatory events
    :param decoder: A dictionary with the decoder to be used in the evaluation of the regulatory events
    :return: A pandas DataFrame with the results of the analysis
    """
    if not interactions:
        interactions = model.yield_interactions()
    else:
        interactions = [model.interactions[interaction] for interaction in interactions]

    if initial_state is None and strategy == 'all':
        warn('Attention! Missing initial state and calculating "all" coefficients may take some time for large models!')

    dfs = []
    for interaction in interactions:

        for coefficient, regulatory_event in interaction.regulatory_events.items():

            if not regulatory_event.is_none:
                df = regulatory_event.truth_table(values=initial_state,
                                                  strategy=strategy,
                                                  coefficient=coefficient,
                                                  operators=operators,
                                                  decoder=decoder)

                df.index = [interaction.target.id] * df.shape[0]

                dfs.append(df)

    df = pd.concat(dfs)
    result_col = df.pop('result')
    df = pd.concat([result_col, df], axis=1)
    return df



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:

    from io import TextIOWrapper

    from mewpy.io.engines.engine import Engine
    from cobra import Model as Cobra_Model
    from reframed import CBModel as Reframed_Model


class Builder:

    def __init__(self):
        """
        Builder base for an engine wrapper of any kind.
        That is, the builder does not provide anything by itself.
        Instead, this class is to be used by the reader and writer, which are just wrappers of engines.

        Wrapping is accomplished using the composition pattern

        Engines, on the other hand, do the hard work of opening, parsing and reading/writing a specific file or model.
        Basically, engines are, actually, the builders of models or files

        """
        self._engine = None

        # useful for the context manager and the read or write methods
        self._context = False

    @property
    def engine(self) -> 'Engine':
        """
        Returns the engine associated with this builder
        :return: engine
        """
        return self._engine

    @property
    def io(self) -> Union[str, 'TextIOWrapper', 'Cobra_Model', 'Reframed_Model']:
        """
        Returns the IO associated with this builder
        :return: IO
        """
        return self.engine.io

    @property
    def config(self) -> dict:
        """
        Returns the configuration associated with this builder
        :return: configuration
        """
        return self.engine.config

    @property
    def warnings(self) -> tuple:
        """
        Returns the warnings associated with this builder
        :return: warnings
        """
        return tuple(self.engine.warnings)



from math import inf
from mewpy.solvers.solver import VarType
from mewpy.solvers import solver_instance
from mewpy.simulation import get_simulator, SStatus
from mewpy.simulation.simulation import Simulator


def epFBA(model, objective=None, reactions=None, constraints=None, obj_frac=None):
    """
    Modified versions of the Parsimonious Flux Balance Analysis allowing to minimize
    the sum of enzyme usage instead of the sum of reaction flux rates when a model includes 
    enzymatic constraints, such as GECKO and sMOMENT formulations.

    If the model defines protein constraints, and no set of reactions are defined,
    the objective will be to minimize enzyme usage, otherwise the objective is to
    minimize the sum of metabolic fluxes.

    :param model: a COBRAPY or REFRAMED model, or an instance of Simulator
    :param objective: The linear objective function as a dict of reaction and coefficient, defaults to None
    :type objective: dict, optional
    :param reactions: list of reactions whose sum of fluxes is to be minimized, defaults to None
    :type reactions: list, optional
    :param constraints: constraints to be imposed, defaults to None
    :type constraints: dict, optional
    :param obj_frac: fraction of the objective, defaults to None
    :type obj_frac: float, optional
    :return: a solver solution
    :rtype: mewpy.solver.Solution
    """

    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    if not objective:
        objective = sim.objective

    solver = solver_instance(sim)

    if not constraints:
        constraints = {}

    # update with simulation constraints if any
    constraints.update(sim.environmental_conditions)
    # constraints.update(sim._constraints)

    # make irreversible
    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_variable(pos, 0, inf, update=False)
            solver.add_variable(neg, 0, inf, update=False)
    solver.update()

    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_constraint(
                'c' + pos, {r_id: -1, pos: 1}, '>', 0, update=False)
            solver.add_constraint(
                'c' + neg, {r_id: 1, neg: 1}, '>', 0, update=False)

    solver.update()

    # add biomass constraint
    pre_solution = sim.simulate(constraints=constraints)

    if pre_solution.status != SStatus.OPTIMAL:
        return pre_solution

    if obj_frac is None:
        solver.add_constraint('obj', objective, '=', pre_solution.objective_value)
    else:
        solver.add_constraint('obj', objective, '>',
                              obj_frac * pre_solution.objective_value)

    solver.update()

    if not reactions:
        try:
            proteins = sim.proteins
            if proteins:
                reactions = [f"{sim.protein_prefix}{protein}" for protein in proteins]
        except Exception:
            reactions = sim.reactions

    sobjective = dict()

    if isinstance(reactions, dict):
        for k, v in reactions.items():
            sobjective[f"{sim.protein_prefix}{k}"] = v
    else:
        for r_id in reactions:
            lb, _ = sim.get_reaction_bounds(r_id)
            if lb < 0:
                pos, neg = r_id + '+', r_id + '-'
                sobjective[pos] = 1
                sobjective[neg] = 1
            else:
                sobjective[r_id] = 1

    solution = solver.solve(sobjective, minimize=True, constraints=constraints)

    return solution




def generation_termination(population, num_generations, num_evaluations, args):
    """Return True if the number of generations meets or exceeds a maximum.

    This function compares the number of generations with a specified 
    maximum. It returns True if the maximum is met or exceeded.

    .. Arguments:
       population -- the population of Individuals
       num_generations -- the number of elapsed generations
       num_evaluations -- the number of candidate solution evaluations
       args -- a dictionary of keyword arguments

    Optional keyword arguments in args:

    - *max_generations* -- the maximum generations (default 1) 

    """
    max_gen = args.get('max_generations', 1)
    if isinstance(max_gen, tuple):
        max_generations = max_gen[0]
    else:
        max_generations = max_gen
    return num_generations >= max_generations
